var request = require('request')
process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
var host = null;
var hypervisor_path = require('./config').hypervisor_path
var hypervisorvm_path = require('./config').hypervisorvm_path
var recoverypointer_path = require('./config').recoverypointer_path
var setSessionandHeader = require('./helpfunction').setSessionandHeader
var ngp = require('../../ecxngp');
const json2csv = require('json2csv').parse;

var option = {
    method:'GET',
    url:'',
    headers:
        {
            'x-endeavour-sessionid':'',
            'x-endeavour-auditrequest':'yes',
            'Content-Type':'application/json',
            'Accept':'application/json'
        },
};
var returnres ={
    "links":{
        "self" :{
            "rel" : "self",
            "href" : ""
        }  
    }
}
var vminfo_list= new Array();
var vmcounts = {};
var hypervisor_addr = {};
var hypervisor_type = {};

function parsefilter(query,next)
{
    try{
        if(typeof(query['filter']) == undefined || query['filter'] ==null ) return {}
        var arr = JSON.parse(query['filter'])
        var filters = {};
        for ( var i in arr )
        {
            if( isNaN(arr[i].value))
            {

                var params = new Set(arr[i].value);
                if( params.has('All') || arr[i].value == "All")
                {
                    continue;
                }
                filters[arr[i].property] = params;
            } 
            else
            {
                filters[arr[i].property] = arr[i].value;
            }
        }
        return filters;
    }
    catch(error)
    {
        console.error(error)
        const err = new Error("Incorrect filtering Parameters")
        err.status = 400
        next(err)
    }
}
function gethypervisor(starturl,resolve,reject)
{
    option.url = starturl
    function loop()
    {
        request( option , function(error, response,body){
            if(!error && response.statusCode == 200)
            {
                var hyperlists = JSON.parse(body).hypervisors
                var links = JSON.parse(body).links
                for( var i in hyperlists )
                {
                    hypervisor_addr[hyperlists[i].id] = hyperlists[i].name;

                    if(hyperlists[i].type == "vmware")
                    hypervisor_type[hyperlists[i].id] = "VMware";
                    else if(hyperlists[i].type == "hyperv")
                    hypervisor_type[hyperlists[i].id] = "Hyper-V";
                    else
                    hypervisor_type[hyperlists[i].id] = hyperlists[i].type;
                }
                if('nextPage' in links)
                {
                    option.url = links.nextPage.href
                    return loop()
                }
                else
                {
                    return resolve()
                } 
            }
            else
            {
                console.error(error)
                const err = new Error("Error when getting hypervisor info")
                err.status = 403
                reject(err)
            }  
        })
    }
    return loop()
}
function Issatisfy(uiItem,vminfo,vmItem,filters)
{
    return !(uiItem in filters) ||  filters[uiItem].has(vminfo[vmItem])
}
function Istimesatisfy(ProtectionTime,filters)
{
    if(!('Backup History' in filters)) return true;
    var currenttime = new Date().getTime();
    var timelimit = currenttime - filters['Backup History']*24*60*60*1000;
    if( ProtectionTime > timelimit ) return true;
    return false;
}
function parseinfo( ori,vminfo,sucess,filters)
{
    vminfo['pk'] = ori['pk'];
    vminfo['VM'] = ori['name'];
    vminfo['tags'] = ori['tags'];
    if(!Issatisfy('VMs',vminfo,'VM',filters)) return false;
    vminfo['SLAPolicy'] = ori.protectionInfo.storageProfileName;
    if(!Issatisfy( 'SLA Policy',vminfo,'SLAPolicy',filters)) return false;
    vminfo['ProtectionTime'] = ori['protectionInfo']['protectionTime'];
    if(!Istimesatisfy(vminfo['ProtectionTime'],filters)) return false;
    vminfo['hypervisor'] = hypervisor_addr[ori.hypervisorKey];
    vminfo['hypervisorType'] = hypervisor_type[ori.hypervisorKey];
    if(sucess)
    {
        vminfo['Status'] = 'Success';
        vminfo['BackupStorage'] = Object.keys(JSON.parse(ori.protectionInfo.storageInfo))[0];
        var transferSize = ori.protectionInfo.transferSize;
        var transferDuration = ori.protectionInfo.transferDuration;
        vminfo['BackupDetails'] ={ 'BackupData' : transferSize , 'BackupDuration' : transferDuration}
    }
    else{
        vminfo['Status'] = 'Failed';
        vminfo['BackupStorage'] = '';
        vminfo['BackupDetails'] = { 'FailureReason' : ori.protectionInfo.lastErrorMessage}
    }
    return true
}

function getlinks(starturl,resolve,reject)
{
    linklists = new Array();
    option.url = starturl
    function loop()
    {
        request(option,function(error,response,body){
            if(!error && response.statusCode == 200 )
            {
                var child_lists = JSON.parse(body).children;
                var links = JSON.parse(body).links;
                for( var index in child_lists)
                {
                   if('versions' in child_lists[index].links){linklists.push(child_lists[index].links.versions.href)};
                }
                if('nextPage' in links)
                {
                    option.url = links.nextPage.href
                    return loop()
                }
                else
                {
                    return resolve(linklists)
                } 
            }
            else if( error )
            {
                console.error(error)
                const err = new Error("Error when getting page links")
                err.status = 403
                reject(err);
            }
        });        
    }
    return loop();
}

function getrecoveryvms(linklists,filters,resolve,reject)
{
    if('Status' in filters && filters['Status'].has('Success')) return resolve(); 
    var promise_array = new Array();
    for( var index in linklists)
    {
        var promise = new Promise( function(resolve,reject){
            option.url = linklists[index];
            function loop()
            {
                request(option,function(error,response,body){
                    if(!error && response.statusCode == 200 )
                    {
                        var contents = JSON.parse(body).contents;
                        var links = JSON.parse(body).links;
                        for( var item in contents )
                        {
                            var vm = contents[item];
                            var vminfo = {};
                            if(!parseinfo(vm.properties.errObj,vminfo,false,filters)) continue;
                            vmcounts[vminfo['VM']] = vminfo['VM'] in vmcounts? vmcounts[vminfo['VM']] + 1 : 1;
                            vminfo_list.push(vminfo);
                        }
                        if('nextPage' in links)
                        {
                            option.url = links.nextPage.href
                             loop()
                        }
                        else
                        {
                             resolve()
                        } 
                        
                    }
                    else if( error )
                    {
                        console.error(error)
                        const err = new Error("Error when getting failed backup history")
                        err.status = 403
                        reject(err);
                    }
                });
            }
            return loop();
        });
        promise_array.push(promise);
    }
    Promise.all(promise_array).then(function()
    {
        resolve();
    }).catch(function(error)
    {
        console.error(error)
        const err = new Error("Error when getting failed backup history")
        err.status = 403
        reject(err);
    })

}

function gethypervisorvm(linklists,filters,resolve,reject){
        if('Status' in filters && filters['Status'].has('Failed')) return resolve(); 
        var promise_array = new Array();
        for( var index in linklists)
        {
            var promise = new Promise( function(resolve,reject){
                option.url = linklists[index];           
                function loop()
                {
                    request(option,function(error,response,body){
                        if(!error && response.statusCode == 200 )
                        {
                            var vmlists = JSON.parse(body).contents;
                            var links = JSON.parse(body).links;
                            for( var index in vmlists)
                            {
                                var vm = vmlists[index]
                                var vminfo = {};
                                if(!parseinfo(vm.properties,vminfo,true,filters)) continue;
                                vmcounts[vminfo['VM']] = vminfo['VM'] in vmcounts? vmcounts[vminfo['VM']] + 1 : 1;
                                vminfo_list.push(vminfo);
                            }
                            if('nextPage' in links)
                            {
                                option.url = links.nextPage.href
                                 loop()
                            }
                            else
                            {
                                 resolve()
                            } 
                        }
                        else if( error )
                        {
                            console.error(error)
                            const err = new Error("Error while getting sucessful backup history")
                            err.status = 403
                            reject(err);
                        }
                    });
                }
                loop()
            });
            promise_array.push(promise);
        }
        Promise.all(promise_array).then(function()
        {
            resolve();
        }).catch(function(error)
        {
            console.error(error)
            const err = new Error("Error while getting sucessful backup history")
            err.status = 403
            reject(err);
        })
}
function compare( a,b )
{
    if((a.VM === b.VM && a.ProtectionTime < b.ProtectionTime) || (a.VM > b.VM))
    {
        return 1;
    }
    else return -1;
}
function countvms(res,req){
    vminfo_list.sort(compare)
    for( var index in vminfo_list )
    {
    var vm = vminfo_list[index]
    vm['TotalRuns'] = vmcounts[vm.VM];
    }
    returnres["instances"] = vminfo_list;
    if('download' in req.query && req.query['download'] == 'true')
    {
        if('type' in req.query && req.query['type'] == 'csv')
        {
            res.setHeader('Content-Disposition', 'attachment;filename = "report.csv"')
            if(vminfo_list.length == 0)
            {
                const fields = ['VMs', 'SLAPolicy', 'ProjectTime','Hypervisor','Status','BackupStorage','BackupDetails'];
                const opts = { fields };    
                var csv = json2csv(vminfo_list,opts);
                res.write(csv)
                res.end()                       
            }
            else
            {
                var csv = json2csv(vminfo_list);
                res.write(csv)
                res.end()   
            }

        }
        else
        {
            res.setHeader('Content-Disposition', 'attachment')
            res.json(vminfo_list)
        }
    }
    else
    {
        res.json(returnres)
    }
}

function getbackuphistory(filters,res,req,next)
{
    setSessionandHeader(res,req)
    .then(function(value){
        option.headers['x-endeavour-sessionid'] = value[0]
        host = value[1];
        returnres["links"]["self"]["href"] = host + "/ngp/analytics/vmbackuphistory";
        vminfo_list = [];
        vmcounts = {};
        hypervisor_addr = {};
    })
    .then(function(){
        return new Promise( function(resolve,reject ){
            gethypervisor(host+hypervisor_path,resolve,reject)
        })
    })
    .then(function(){
        return new Promise(function(resolve,reject){
            getlinks(host+hypervisorvm_path,resolve,reject)
        })
    })
    .then( function(linklists){
        return new Promise( function(resolve,reject){
            gethypervisorvm(linklists,filters,resolve,reject)
        })
    })
    .then( function()
    {
        return new Promise( function(resolve,reject){
            getlinks(host+recoverypointer_path,resolve,reject)
        })
    })
    .then(function(linklists){
        return new Promise( function( resolve,reject){
            getrecoveryvms(linklists,filters,resolve,reject)
        })
    })
    .then(function()
    {
        countvms(res,req)
    })
    .catch(function(error){
        console.log(error)
        next(error)
    })  
}

module.exports =
    {
        getbackuphistory : getbackuphistory,
        parsefilter : parsefilter
    };