if (!('contains' in String.prototype)) {
    String.prototype.contains = function(str, startIndex) {
      return ''.indexOf.call(this, str, startIndex) !== -1;
    };
  }

if (typeof String.prototype.startsWith != 'function') {
  String.prototype.startsWith = function(prefix) {
      return this.slice(0, prefix.length) == prefix;
  };
}
       
if (typeof String.prototype.endsWith != 'function') {
  String.prototype.endsWith = function(suffix) {
      return this.slice(-suffix.length) == suffix;
  };
}


var ngp = require('./ecxngp');                                                                

/*
    All functions GET, DELETE, PUT related to the lifecycle management of restore jobs are housed here. Consequently, the customary sequence of functional steps must be modified
    because, depending on what verb/function is in play, a completely different set of steps may be required. So define step streams for each of the 3 functions and make effective
    a given step stream based on the verb/function.
*/

var funcStepsGet = [                                                                             // for GET /ecxngp/application/job?subtype=vmware|hyperv
  {                                                                                              
      continue: 'next',                                                                          
      name: 'get_restorejobs',
      verb: 'GET',
      path: '/api/endeavour/job?filter=[{"property":"subType","value":"qqVVqq"},{"property":"type","value":"recovery"},ONDEMAND_FILTER]', 
      ip: null,                                                                                                                                       
      port: null, 
      type: null,
      func: null,                                                                                
      cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
  }, 
];

var funcStepsDelete = [                                                                          // for DELETE /ecxngp/application/job?name=<job_name>
    {                                                                                              
        continue: 'next',                                                                          
        name: 'del_restorejob_getpolicy',
        verb: 'GET',
        path: '/api/endeavour/policy?filter=[{"property":"name","value":"qqVVqq"}]', 
        ip: null,                                                                                                                                     
        port: null, 
        type: null,
        func: null,                                                                                 
        cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
    },     
  ];

  var funcStepsPut = [                                                                           // for PUT /ecxngp/application/job?action=restore
    {                                                                                              
        continue: 'next',                                                                          
        name: 'put_restorejob',
        verb: 'GET',
        path: '/api/endeavour/policy?filter=[{"property":"name","value":"qqVVqq"}]', 
        ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
        port: null, 
        type: null,
        func: null,                                                                                // see initialization below 
        cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
    }, 
  ];  


const getJobSessionStepFuncs = {
  get_restorejobs: function(step, funcStream, ecxSessionId, oFromNg, cb){                            
    getRestoreJobs(step, funcStream, ecxSessionId, oFromNg, cb);   
  },
  put_restorejob: function(step, funcStream, ecxSessionId, oFromNg, cb){                            
    putRestoreJob(step, funcStream, ecxSessionId, oFromNg, cb);   
  },  
  del_restorejob_getpolicy: function(step, funcStream, ecxSessionId, oFromNg, cb){                            
    getRestorePolicy(step, funcStream, ecxSessionId, oFromNg, cb);   
  },  
};


exports.init = init;


// query string (keyVals) varies by verb: get, put, delete. I intend to house all 3 functions in this module - so must discriminate 

function init(req, res, keyVals)                                                    
{
    var ngState = ngp.getSessionState(req, res);
    
    if ( ngState.stale )                                                                           // FTI we got bounced while client remained up. AND response already sent.
        return;                                                                                    // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

    var method = null;
    var funcSteps = null;
    var genericFuncName = null;

    if ( req.method == 'GET' )
    {
        method = 'get';
        genericFuncName = 'get_restorejobs';
        funcSteps = JSON.parse(JSON.stringify(funcStepsGet));
    }

    if ( req.method == 'DELETE' )
    {
        method = 'delete';
        genericFuncName = 'delete_restorejob';
        funcSteps = JSON.parse(JSON.stringify(funcStepsDelete));
    }    

    if ( req.method == 'PUT' )
    {
        method = 'put';
        genericFuncName = 'upd_restorejob';
        funcSteps = JSON.parse(JSON.stringify(funcStepsPut));
    }   

    var ecxSid = ngState.ecxSid;
    res.ecxSid = ecxSid;

    var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

    stepsStream.forEach(function(step){
        step.ip = ngp.localIp;
        step.port = ngp.ecxExternal;

        if ( step.name == 'get_restorejobs'  )
        {
            step.path = step.path.replace('qqVVqq', keyVals.subtype);

            if ( req.url.contains('&') )
            {
                var filter = req.url.split('&')[1];
                var escFilter = require('querystring').unescape(filter);

                //escFilter = escFilter.replace('operator', 'op');    //!!! HACK, bypass gui bug

                escFilter = escFilter.split('=')[1];
                escFilter = escFilter.replace('[', '');
                escFilter = escFilter.replace(']', ''); 

                step.path = step.path.replace('ONDEMAND_FILTER', escFilter);           // fO[0] is the "tuple" (thank you, Erlang) inside the filter array
            }
            else
            {
                step.path = step.path.replace(',ONDEMAND_FILTER', '');
            }
         }

        if ( step.name == 'del_restorejob_getpolicy' )
        {
            step.path = step.path.replace('qqVVqq', keyVals.name);
        }

        step.func = getJobSessionStepFuncs[step.name];                                             // hook up the step's real workhorse
        
        step.cb = function(ecxSid, data){ 
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

            if ( data )
            {
                ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
            }
        };
    });

    ngp.createRequestStore(stepsStream, ecxSid, req.reqId, genericFuncName);  
    
    ngState.currentFunc = genericFuncName;                                                         
    return stepsStream;
}


var getRestorePolicy = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
  var ngState = ngp.ecxSidHash[ecxSessionId];

  var entityIdsToDelete = [];

  var path = step.path;    
  
  var currentStepPathSplitForFilter = path.split('=');
  var currentStepPathFilter = currentStepPathSplitForFilter[1];
  
  var fixedFilterOption = ngp.escape(currentStepPathFilter);
  var fixedStepPath = currentStepPathSplitForFilter[0] + '=' + fixedFilterOption;

  ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(rstPolO){ 
      if ( rstPolO.e || rstPolO.response.statusCode != 200 )  
      {
          var failResult = ngp.getGenericEcxErrorResponse(rstPolO, ecxSessionId, step); 
          ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
      }
      else
      {
          if ( rstPolO.rO.policies.length < 1 )                                                    // no policy with this name...
          {
              var notFoundResult = {
                  statusCode: 404,
                  response: null,
              };

              ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, notFoundResult, false);
              return;
          }

          var rstPolicy = rstPolO.rO.policies[0];


          var path = step.path.replace('vour/policy', 'vour/job');  
          
          var currentStepPathSplitForFilter = path.split('=');
          var currentStepPathFilter = currentStepPathSplitForFilter[1];
          
          var fixedFilterOption = ngp.escape(currentStepPathFilter);
          var fixedStepPath = currentStepPathSplitForFilter[0] + '=' + fixedFilterOption;
        
        
          ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(rstJobO){
            var rstJob = rstJobO.rO.jobs[0];


            entityIdsToDelete.unshift({trigger: rstJob.triggerIds[0]});                            // delete trigger last
            entityIdsToDelete.unshift({policy: rstPolicy.id});                                     // delete policy 2nd
            entityIdsToDelete.unshift({job: rstJob.id});                                           // delete job 1st

            deleteRestoreJobEntities(entityIdsToDelete, ecxSessionId, function(rc, result){
                ngp.logMsg('warn', 'DELETION DONE');
                
                var successResult = {
                    statusCode: 201,
                    response: {thisIsATest: true},
                };

                ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
            });
          });
      }
  });
};


function deleteRestoreJobEntities(entityIdsToDelete, ecxSessionId, cb)
{
    if ( entityIdsToDelete.length < 1 )
    {
        cb('ok', null);
    }
    else
    {
        var eid = entityIdsToDelete.shift();

        for ( entityName in eid )
            var entityId = entityName;
        
        var path = '/api/endeavour/' + entityId + '/' + eid[entityId];

        ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, path, ecxSessionId, null, function(delO){  // works or it don't, SRTE will have complained
            /*
            if ( delO.e || ( delO.response.statusCode != 204 && delO.response.statusCode != 404 ) ) 
            {
                ngp.logMsg('warn', 'SOME DELETE PROBLEM=' + JSON.stringify(delO.rO));
            }
            */
            
            deleteRestoreJobEntities(entityIdsToDelete, ecxSessionId, cb);
        });
    }
}


var getRestoreJobs = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
  var ngState = ngp.ecxSidHash[ecxSessionId];

  var currentStepPath = step.path;
  var currentStepPathSplitForFilter = currentStepPath.split('=');
  var currentStepPathFilter = currentStepPathSplitForFilter[1];
  
  var fixedFilterOption = ngp.escape(currentStepPathFilter);
  var fixedStepPath = currentStepPathSplitForFilter[0] + '=' + fixedFilterOption;
  
  ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(rstJobsO){ 
      if ( rstJobsO.e || rstJobsO.response.statusCode != 200 )  
      {
          var failResult = ngp.getGenericEcxErrorResponse(rstJobsO, ecxSessionId, step); 
          ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
      }
      else
      {
          //ngp.logMsg('warn', "Get restore jobs OK response=" + JSON.stringify(rstJobsO.rO));  

          var restoreJobs = rstJobsO.rO.jobs;       
          var candidateRestoreJobs = [];

          restoreJobs.forEach(function(restoreJob){
              if ( ! restoreJob.name.startsWith('onDemand_Restore') )
                candidateRestoreJobs.push(restoreJob);
          });

          if ( candidateRestoreJobs.length < 1 )
          {
              rstJobsO.rO.jobs = [];
              rstJobsO.rO.total = 0;

              var successResult = {
                  statusCode: 201,
                  response: rstJobsO.rO,
              };

              ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
          }
          else
          {
              var policyModifiedCandidates = [];

              getRestoreJobPolicies(candidateRestoreJobs, policyModifiedCandidates, step, ecxSessionId, function(rc, result){
                    if ( rc != 'ok' )
                    {
                        ngp.logMsg('warn', 'POLICY ' + rc + '=' + JSON.stringify(result));

                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, result, false);  
                        return;
                    }
                    
                    var modifiedCandidates = [];

                    getRestoreJobTriggers(policyModifiedCandidates, modifiedCandidates, step, ecxSessionId, function(rc, result){
                        if ( rc != 'ok' )
                        {
                            ngp.logMsg('warn', 'TRIGGER ' + rc + '=' + JSON.stringify(result));
                            
                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, result, false);  
                            return;
                        }
                        
                        rstJobsO.rO.jobs = modifiedCandidates;
                        rstJobsO.rO.total = modifiedCandidates.length;
    
                        var successResult = {
                            statusCode: 200,
                            response:rstJobsO.rO,
                        };
        
                        ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
                    });
              });
          }
      }
  });
};


var KEEP_getRestoreJobs = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
  var ngState = ngp.ecxSidHash[ecxSessionId];

  ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, null, function(rstJobsO){ 
      if ( rstJobsO.e || rstJobsO.response.statusCode != 200 )  
      {
          var failResult = ngp.getGenericEcxErrorResponse(rstJobsO, ecxSessionId, step); 
          ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
      }
      else
      {
          //ngp.logMsg('warn', "Get restore jobs OK response=" + JSON.stringify(rstJobsO.rO));  

          var restoreJobs = rstJobsO.rO.jobs;       
          var candidateRestoreJobs = [];

          restoreJobs.forEach(function(restoreJob){
              if ( ! restoreJob.name.startsWith('onDemand_Restore') )
                candidateRestoreJobs.push(restoreJob);
          });

          if ( candidateRestoreJobs.length < 1 )
          {
              rstJobsO.rO.jobs = [];
              rstJobsO.rO.total = 0;

              var successResult = {
                  statusCode: 201,
                  response: rstJobsO.rO,
              };

              ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
          }
          else
          {
              var modifiedCandidates = [];
              getRestoreJobTriggers(candidateRestoreJobs, modifiedCandidates, step, ecxSessionId, function(rc, result){
                  ngp.logMsg('warn', rc + '=' + JSON.stringify(result));
                  
                  rstJobsO.rO.jobs = modifiedCandidates;
                  rstJobsO.rO.total = modifiedCandidates.length;

                  var successResult = {
                    statusCode: 201,
                    response:rstJobsO.rO,
                };
  
                ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
              });
          }
      }
  });
};


var putRestoreJob = function(step, funcStream, ecxSessionId, oFromNg, cb)                               // this cb is step callback
{
  var ngState = ngp.ecxSidHash[ecxSessionId];

  step.path = step.path.replace('qqVVqq', oFromNg.name);  

  var currentStepPath = step.path;
  var currentStepPathSplitForFilter = currentStepPath.split('=');
  var currentStepPathFilter = currentStepPathSplitForFilter[1];
  
  var fixedFilterOption = ngp.escape(currentStepPathFilter);
  var fixedStepPath = currentStepPathSplitForFilter[0] + '=' + fixedFilterOption;
  
  ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(rstPolO){    // get summary (inside policies array) object
      if ( rstPolO.e || rstPolO.response.statusCode != 200 )  
      {
          var failResult = ngp.getGenericEcxErrorResponse(rstPolO, ecxSessionId, step); 
          ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
      }
      else
      {
         // ngp.logMsg('warn', "Get restore policy for PUT OK response=" + JSON.stringify(rstPolO.rO)); 

          var policiesArray = rstPolO.rO.policies;

          if ( policiesArray.length < 1 )
          {
              var failResult = {
                  statusCode: 404,
                  response: null,
              };

              ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);  
              return;
          }

          var path = policiesArray[0].links.edit.href;                                                  // summary object in policies array lacks trigger info, so get real thing

            ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, path, ecxSessionId, null, function(rstPolDtlO){
            if ( rstPolDtlO.e || rstPolDtlO.response.statusCode != 200 )  
                {
                    var failResult = ngp.getGenericEcxErrorResponse(rstPolDtlO, ecxSessionId, step); 
                    ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
                }
                else
                {
                    //@@var triggerFromNg = JSON.parse(JSON.stringify(oFromNg.trigger));                    // Use below in updating extant trigger object..
                    //@@delete oFromNg.trigger;                                                             // ..but don't present to PUT of policy

                    var appResPolCommonTpl = JSON.parse(JSON.stringify(ngp.applicationRestorePolicyTemplate));                    // this is common part of application recovery POST body
                    appResPolCommonTpl.name = oFromNg.name;
                    appResPolCommonTpl.subType = oFromNg.subType;
                    appResPolCommonTpl.spec = oFromNg.spec;
                    appResPolCommonTpl.spec.subpolicy[0].name = oFromNg.name;                              //!! not sure this makes any sense at all, no such source property methinks
                    appResPolCommonTpl.script = oFromNg.script;                    

                    ngp.sendRequestToEcx('PUT', ngp.ecxExternal, path, ecxSessionId, appResPolCommonTpl, function(updPolDtlO){    // Now update the policy
                        if ( updPolDtlO.e || updPolDtlO.response.statusCode != 200 )                          
                        {
                            var failResult = ngp.getGenericEcxErrorResponse(updPolDtlO, ecxSessionId, step); 
                            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
                        }
                        else                                                                            // policy updated, now update its trigger. WHAT IF IT FAILS? ROLLBACK?
                        {
                            //@@getAndUpdateTrigger(step, oFromNg, ngState, triggerFromNg, function(rc, result){      // note that we are 3 CALLBACKS DEEP at this point
                            getAndUpdateTrigger(step, oFromNg, ngState, function(rc, result){           // note that we are 3 callbacks deep at this point                                
                                ngp.logMsg('warn', 'RC=' + rc + ', Result=' + JSON.stringify(result));
                                //process.exit();
                                var jiveResult = {
                                    statusCode: 200,
                                    response: {thisIsATest: true},
                                };
                    
                                ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, jiveResult);                        
                            });
                        }
                    });
                }
          });
      }
  });
};

function getAndUpdateTrigger(step, oFromNg, ngState, cb)
{
    var triggerName = oFromNg.name;                                                                      // and that of policy and job

    ngp.sendRequestToEcx('GET', ngp.ecxExternal, '/api/endeavour/trigger', ngState.ecxSid, null, function(trigO){ 
        if ( trigO.e || trigO.response.statusCode != 200 )  
        {
            var failResult = ngp.getGenericEcxErrorResponse(trigO, ngState.ecxSid, step); 
            cb('ng', failResult);
        }
        else                                                                                             // EHI 0 or more trigger in the triggers array (FILTER DON'T WORK)      
        {
            var foundTrigger = false;

            var triggers = trigO.rO.triggers;                                                            // exactly 1 trigger in the array

            triggers.forEach(function(trigger){
                if ( foundTrigger )
                    return;

                if ( trigger.name == triggerName )
                {
                    foundTrigger = trigger;
                    return;
                }
            });

            if ( ! foundTrigger )
            {
                var failResult = {
                    statusCode: 404,
                    response: {rc: 'Failed to find trigger for policy/job ' + triggerName},     
                };

                cb('ng', failResult);
                return;
            }

            foundTrigger.properties = oFromNg.trigger;
            delete foundTrigger.links;
            delete foundTrigger.rbacPath;
            delete foundTrigger.onHold;
            var triggerId = foundTrigger.id;
            delete foundTrigger.id;

            // foundTrigger.properties.activateDate = (new Date).getTime();  //!!! PROBABLY WRONG; vide ECC-11717
            foundTrigger.properties.activateDate = oFromNg.trigger.activateDate;    // ECC-11717

            ngp.sendRequestToEcx('PUT', ngp.ecxExternal, '/api/endeavour/trigger/' + triggerId, ngState.ecxSid, foundTrigger, function(putTrigO){
                if ( putTrigO.e || putTrigO.response.statusCode != 200 )  
                    {
                        var failResult = ngp.getGenericEcxErrorResponse(putTrigO, ngState.ecxSid, step); 
                        cb('ng', failResult);
                    }
                    else
                    {
                        cb('ok', null);
                    }
            });
        }
    }); 
}

function LAST_getAndUpdateTrigger(step, oFromNg, ngState, triggerFromNg, cb)
{
    var triggerName = oFromNg.name;                                                                      // and that of policy and job

    ngp.sendRequestToEcx('GET', ngp.ecxExternal, '/api/endeavour/trigger', ngState.ecxSid, null, function(trigO){ 
        if ( trigO.e || trigO.response.statusCode != 200 )  
        {
            var failResult = ngp.getGenericEcxErrorResponse(trigO, ngState.ecxSid, step); 
            cb('ng', failResult);
        }
        else                                                                                             // EHI 0 or more trigger in the triggers array (FILTER DON'T WORK)      
        {
            var foundTrigger = false;

            var triggers = trigO.rO.triggers;                                                            // exactly 1 trigger in the array

            triggers.forEach(function(trigger){
                if ( foundTrigger )
                    return;

                if ( trigger.name == triggerName )
                {
                    foundTrigger = trigger;
                    return;
                }
            });

            if ( ! foundTrigger )
            {
                var failResult = {
                    statusCode: 404,
                    response: {rc: 'Failed to find trigger for policy/job ' + triggerName},     
                };

                cb('ng', failResult);
                return;
            }

            foundTrigger.properties = triggerFromNg;
            //foundTrigger.properties.activateDate = (new Date).getTime();
            //foundTrigger.properties.deactivateDate = null;

            ngp.sendRequestToEcx('PUT', ngp.ecxExternal, '/api/endeavour/trigger/' + foundTrigger.id, ngState.ecxSid, foundTrigger, function(putTrigO){
                if ( putTrigO.e || putTrigO.response.statusCode != 200 )  
                    {
                        var failResult = ngp.getGenericEcxErrorResponse(putTrigO, ngState.ecxSid, step); 
                        cb('ng', failResult);
                    }
                    else
                    {
                        cb('ok', null);
                    }
            });
        }
    }); 
}

function getRestoreJobTriggers(candidateRestoreJobs, modifiedCandidates, step, ecxSessionId, cb)
{
    if ( candidateRestoreJobs.length < 1 )
    {
        cb('ok', modifiedCandidates);
    }
    else
    {
        var crj = candidateRestoreJobs.shift();
        ngp.sendRequestToEcx('GET', ngp.ecxExternal, crj.links.triggers.href, ecxSessionId, null, function(triggerO){
            if ( triggerO.e || triggerO.response.statusCode != 200 )  
            {
                var failResult = ngp.getGenericEcxErrorResponse(triggerO, ecxSessionId, step); 
                cb('ng', failResult);
            }
            else 
            {
                crj.trigger = {};

                if ( triggerO.rO.triggers.length > 0 )
                {
                    var trigger = triggerO.rO.triggers[0];                                             // exactly 1 trigger with top-level property 'properties', or maybe not
                    crj.trigger = trigger.properties;
                }
                
                modifiedCandidates.push(crj);
                getRestoreJobTriggers(candidateRestoreJobs, modifiedCandidates, step, ecxSessionId, cb);    // RECURSE
            }
        });
    }
}

function getRestoreJobPolicies(candidateRestoreJobs, modifiedCandidates, step, ecxSessionId, cb)
{
    if ( candidateRestoreJobs.length < 1 )
    {
        cb('ok', modifiedCandidates);
    }
    else
    {
        var crj = candidateRestoreJobs.shift();

        ngp.sendRequestToEcx('GET', ngp.ecxExternal, crj.links.policyedit.href, ecxSessionId, null, function(polO){
            if ( polO.e || polO.response.statusCode != 200 )  
            {
                var failResult = ngp.getGenericEcxErrorResponse(polO, ecxSessionId, step); 
                cb('ng', failResult);
            }
            else 
            {
                crj.spec = {};
                crj.script = {};
                
                if ( polO.rO.spec )
                    crj.spec = polO.rO.spec;
                if ( polO.rO.script )
                    crj.script = polO.rO.script;
                
                modifiedCandidates.push(crj);
                getRestoreJobPolicies(candidateRestoreJobs, modifiedCandidates, step, ecxSessionId, cb);    // RECURSE
            }
        });
    }
}