if (!('contains' in String.prototype)) {
    String.prototype.contains = function(str, startIndex) {
      return ''.indexOf.call(this, str, startIndex) !== -1;
    };
  }

if (typeof String.prototype.startsWith != 'function') {
  String.prototype.startsWith = function(prefix) {
      return this.slice(0, prefix.length) == prefix;
  };
}

if (typeof String.prototype.endsWith != 'function') {
  String.prototype.endsWith = function(suffix) {
      return this.slice(-suffix.length) == suffix;
  };
}

var ngp = require('../ecxngp');
var url = require('url');

const catalogRecoveryTplFile = ngp.PUBLICDIR + '/catalogrpoltpl.json';
var catalogRecoveryTpl = null;

var catalogRestoreFuncSteps = [
	{
		continue: 'next',
		name: 'catalog_recovery_mode',
		verb: null,
		path: null,
		ip: null,
		port: null,
		type: null,
		func: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'catalog_check_recovery_policy_exists',
		verb: null,
		path: null,
		ip: null,
		port: null,
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'catalog_recovery_check_and_update_trigger',
		verb: null,
		path: null,
		ip: null,
		port: null,
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'catalog_recovery_create_trigger',
		verb: null,
		path: null,
		ip: null,
		port: null,
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'catalog_recovery_create_policy',
		verb: null,
		path: null,
		ip: null,
		port: null,
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
	{
		continue: 'next',
		name: 'catalog_recovery_rollback',
		verb: null,
		path: null,
		ip: null,
		port: null,
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    }
];

const catalogRestoreFuncs = {
    //determines if it is an onDemand (or) scheduled mode
    catalog_recovery_mode: function(step, funcStream, ecxSessionId, oFromNg, cb){
		ngp.recoveryMode(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    //check for scheduled mode
	catalog_check_recovery_policy_exists: function(step, funcStream, ecxSessionId, oFromNg, cb){
        checkCatalogRecoveryPolicyExists(step, funcStream, ecxSessionId, oFromNg, cb);
    },
	catalog_recovery_check_and_update_trigger: function(step, funcStream, ecxSessionId, oFromNg, cb){
		catalogRestoreCheckAndUpdateTrigger(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    catalog_recovery_create_trigger: function(step, funcStream, ecxSessionId, oFromNg, cb){
        catalogRestoreCreateTrigger(step, funcStream, ecxSessionId, oFromNg, cb);
    },
    catalog_recovery_create_policy: function(step, funcStream, ecxSessionId, oFromNg, cb){
        catalogRestoreCreatePolicyJob(step, funcStream, ecxSessionId, oFromNg, cb);
    },
	catalog_recovery_rollback: function(step, funcStream, ecxSessionId, oFromNg, cb){
		catalogRecoveryRollback(step, funcStream, ecxSessionId, oFromNg, cb);
    },
}

function getConfiguration()
{
    try{
        var catalogRecoveryTemp = ngp.fs.readFileSync(catalogRecoveryTplFile, 'utf8');
        catalogRecoveryTpl = JSON.parse(catalogRecoveryTemp);
    }catch(e){
        ngp.logMsg('error', " Failed to read catalog recovery template file because of error = " + e.toString());
        return;
    }
}

/**
 * Checks if recovery policy exists for catalog restore
 *
 * @param  step
 * @param  funcStream
 * @param  ecxSessionId
 * @param  oFromNg
 * @param  cb
 */
function checkCatalogRecoveryPolicyExists(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.addDoubleSpace();
    ngp.logMsg('info', '	Checking for existence of catalog recovery policy [Scheduled Mode] ==>');
    ngp.addDoubleSpace();

    ngp.ngpCoreHelper.getRecoveryPolicies(ecxSessionId, oFromNg, 'catalog', function(getRecoveryPolicesResObj){

        var failResult = null;
		if ( getRecoveryPolicesResObj.e || getRecoveryPolicesResObj.response.statusCode != 200 ){
			failResult = ngp.getGenericEcxErrorResponse(getRecoveryPolicesResObj, ecxSessionId, step);
			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);
		}else{
			var ngState = ngp.ecxSidHash[ecxSessionId];
            var recoveryPolicyExists = false;

            getRecoveryPolicesResObj.rO['policies'].forEach(function(policy){
                if ( recoveryPolicyExists )
                    return;

                if ( policy.name == oFromNg['name'] ){
                    recoveryPolicyExists = true;
                    return;
                }
            });

            if (recoveryPolicyExists){
                failResult = {
                    statusCode: 400,
                    response: {rc: "        Catalog recovery policy [" + oFromNg.name + "] already exists"},
                };

                ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
            }else{
                step.cb(ecxSessionId, null);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
                return;
            }
        }
    });
}

/**
 * Checks if trigger exists and update trigger
 *  a. If the client's POST body contains no .trigger property, then cause the trigger creation step to be skipped and drive the step logic. Otherwise:
 *  b. GET from SPP the triggers array containing, perhaps, the triggers belonging to this restore job
 *  c. If that array is empty, then create the trigger in keeping with the intent expressed by the presence of oFromNg.trigger
 *  d. If that array is not empty (it contains exactly 1 trigger), then update it.
 *
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function catalogRestoreCheckAndUpdateTrigger(step, funcStream, ecxSessionId, oFromNg, cb)
{
    ngp.addDoubleSpace();
    ngp.logMsg('info', '	Checking and updating trigger for catalog recovery policy ==>');
    ngp.addDoubleSpace();

    if (!oFromNg['trigger']){
        if ( funcStream.length > 0 )
                funcStream[0]['continue'] = 'skip';

        step.cb(ecxSessionId, null);
        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
        return;
    }

    ngp.ngpCoreHelper.getTrigger(ecxSessionId, oFromNg['name'], function(getTriggerResObj){

        if ( getTriggerResObj.e || (getTriggerResObj.response.statusCode != 200 && getTriggerResObj.response.statusCode != 404) ){
            var failResult = ngp.getGenericEcxErrorResponse(getTriggerResObj, ecxSessionId, step);
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
        }else if ( getTriggerResObj.rO['triggers'].length < 1 ){
                ngp.logMsg('warn', '        No trigger found for job [' + oFromNg['name'] + ']');
                step.cb(ecxSessionId, null);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
        }else{
            ngp.logMsg('debug', '       Found trigger. Updating trigger for job [' + oFromNg['name'] + ']');

            var triggerPutBody = {
                name: oFromNg['name'],
                type: "Schedule",
                properties: oFromNg['trigger'],
            };

            ngp.ngpCoreHelper.updateTrigger(ecxSessionId, getTriggerResObj.rO['triggers'][0]['id'], triggerPutBody, function(updateTriggerResObj){

                if ( updateTriggerResObj.e || updateTriggerResObj.response.statusCode != 200 ){
                        var failResult = ngp.getGenericEcxErrorResponse(updateTriggerResObj, ecxSessionId, step);
                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                }else{
                        // SKIPPING TRIGGER CREATE STEP
                        if ( funcStream.length > 0 )
                                funcStream[0]['continue'] = 'skip';

                        step.cb(ecxSessionId, null);
                        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
                }
            });
        }
    });
}

/**
 * Create trigger for catalog restore
 *
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function catalogRestoreCreateTrigger(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.addDoubleSpace();
    ngp.logMsg('info', '	Creating trigger for catalog recovery policy ==>');
    ngp.addDoubleSpace();

	var rMode = ngp.getRequestState(step.reqId, ecxSessionId, 'rMode');

    if ( !rMode || !rMode['scheduled'] ){
        // No need to create schedule (!rMode.scheduled ->an onDemand recovery is in play)
        step.cb(ecxSessionId, null);
        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
    }else{
        var triggerTplClone = JSON.parse(JSON.stringify(ngp.triggerTpl));
        triggerTplClone['name'] = oFromNg['name'];
        triggerTplClone['properties'] = oFromNg['trigger'];
        triggerTplClone['properties']['activateDate'] = oFromNg['trigger']['activateDate'];
        triggerTplClone['properties']['deactivateDate'] = null;

        ngp.ngpCoreHelper.createTrigger(ecxSessionId, triggerTplClone, function(createTriggerResObj){
            if ( createTriggerResObj.e || createTriggerResObj.response.statusCode != 201 ){
                var failResult = ngp.getGenericEcxErrorResponse(createTriggerResObj, ecxSessionId, step);
                ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
            }else{
                ngp.setRequestState(step.reqId, ecxSessionId, 'triggerId', createTriggerResObj.rO['id']);
                ngp.setRequestState(step.reqId, ecxSessionId, 'trigger', createTriggerResObj.rO);

                step.cb(ecxSessionId, null);
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
            }
        });
    }
}

function  catalogRestoreCreatePolicyJob(step, funcStream, ecxSessionId, oFromNg, cb){
    ngp.addDoubleSpace();
    ngp.logMsg('info', '    Creating catalog recovery policy and job ==>');
    ngp.addDoubleSpace();

    getConfiguration();

    var crPolicySteps = [];

    var rollback = {
        policies: [],
        jobs: [],
    };

    var triggerId = ngp.getRequestState(step.reqId, ecxSessionId, 'triggerId');

    var rMode = ngp.getRequestState(step.reqId, ecxSessionId, 'rMode');

    if (! rMode['scheduled'])
    {
        oFromNg['name'] = 'onDemandRestore_' + (new Date).getTime();
    }

    var catalogRecoveryPolCommonTpl = JSON.parse(JSON.stringify(catalogRecoveryTpl));
    catalogRecoveryPolCommonTpl['name'] = oFromNg['name'];
    catalogRecoveryPolCommonTpl['subType'] = oFromNg['subType'];
    catalogRecoveryPolCommonTpl['spec'] = oFromNg['spec'];
	catalogRecoveryPolCommonTpl['spec']['subpolicy'][0]['name'] = oFromNg['name'];
	catalogRecoveryPolCommonTpl['script'] = oFromNg['script'];

    var pjEntry = {
        policyName: oFromNg.name,
        policyCreated: false,
        jobCreated: false,
        triggerId: triggerId,
		policyPostBody: catalogRecoveryPolCommonTpl,
        policyRollbackAttempted: false,
        policyRollbackSuccessful: false,
        jobEntity: null,
        jobRollbackAttempted: false,
		jobRollbackSuccessful: false,
		triggerRollbackAttempted: false,
		triggerRollbackSuccessful: false,
		entitiesCreated: 0,
    };

    var crSummaryResults = [];

	if (triggerId)
        crSummaryResults.push(ngp.getRequestState(step.reqId, ecxSessionId, 'trigger'));

    createCatalogRecoveryPolicyAndJob(pjEntry, step, ecxSessionId, crSummaryResults, triggerId, function(){
        if ( pjEntry.entitiesCreated != 2 ){
            ngp.logMsg('error', '        Detected failure of policy and job creation in session [' + ecxSessionId + ']; attempting rollback');
            ngp.setRequestState(step.reqId, ecxSessionId, 'rollbackState', pjEntry);
            step.cb(ecxSessionId, null);
            ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
        }else{
            ngp.logMsg('info', '        Created catalog recovery policy and job');
            if ( funcStream.length > 0 )
                //SKIP ROLLBACK STEP
                funcStream[0]['continue'] = 'skip';

            if (rMode['scheduled']){
                ngp.logMsg('info', '            Catalog recovery job has been scheduled');
                var successResult = {
                    statusCode: 201,
                    response: crSummaryResults,
                };

                ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
            }else{

                var startLink = pjEntry['jobEntity']['links']['start']['href'];
                ngp.logMsg('info', '            Starting catalog recovery job for onDemand mode startlink [' + startLink + ']');

                ngp.sendRequestToEcx('POST', ngp.ecxExternal, startLink, ecxSessionId, null, function(startResponseObj){
                    if ( startResponseObj.e || ( startResponseObj.response.statusCode != 201 && startResponseObj.response.statusCode != 200 )){
                        ngp.logMsg('error', '           Detected failure in starting catalog recovery job for onDemand mode');
                        var failResult = ngp.getGenericEcxErrorResponse(startResponseObj, ecxSessionId, step);
                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
                    }else{
                        ngp.logMsg('info', '            Started catalog recovery job for onDemand mode startlink [' + startLink + ']');
                        var successResult = {
                            statusCode: 201,
                            response: startResponseObj.rO,
                        };

                        ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);
                    }
                });
            }
        }
    });

}

/**
 * Creates catalog recovery policy and job
 *
 * @param crPolicyStep
 * @param driverStep
 * @param ecxSessionId
 * @param crSummaryResults
 * @param triggerId
 * @param catalogPolicyJobCreateCb
 */
function createCatalogRecoveryPolicyAndJob(crPolicyStep, driverStep, ecxSessionId, crSummaryResults, triggerId, catalogPolicyJobCreateCb){

    ngp.ngpCoreHelper.createPolicy(ecxSessionId, crPolicyStep['policyPostBody'], function(createPolicyResObj){
		if ( createPolicyResObj.e || createPolicyResObj.response.statusCode != 201 ){
			var failResult = ngp.getGenericEcxErrorResponse(createPolicyResObj, ecxSessionId, driverStep);
			catalogPolicyJobCreateCb(failResult);
		}else{            
			crSummaryResults.push(createPolicyResObj.rO);
            crPolicyStep['policyId'] = createPolicyResObj.rO['id'];
            crPolicyStep['policyCreated'] = true;
            crPolicyStep.entitiesCreated++;

            ngp.logMsg('info', '        Created catalog recovery policy = ' + JSON.stringify(createPolicyResObj.rO));

            var jobTplClone = JSON.parse(JSON.stringify(ngp.jobTpl));
            jobTplClone['name'] = crPolicyStep['policyName'];
            jobTplClone['description'] += crPolicyStep['policyName'];
            jobTplClone['policyId'] = createPolicyResObj.rO['id'];

            if ( triggerId )
                jobTplClone['triggerIds'].push(triggerId);

            ngp.ngpCoreHelper.createJob(ecxSessionId, jobTplClone, function(createJobResObj){
				if ( createJobResObj.e || createJobResObj.response.statusCode != 201 ){                    
					var failResult = ngp.getGenericEcxErrorResponse(createJobResObj, ecxSessionId, driverStep);
					catalogPolicyJobCreateCb(failResult);
				}else{
					crSummaryResults.push(createJobResObj.rO);
					crPolicyStep['jobEntity'] = createJobResObj.rO;
					crPolicyStep['jobId'] = createJobResObj.rO['id'];
					crPolicyStep['jobCreated'] = true;
					crPolicyStep.entitiesCreated++;

					ngp.logMsg('info', '        Created catalog recovery job = ' + JSON.stringify(createJobResObj.rO));

					catalogPolicyJobCreateCb();
				}
            });

        }
    });
}

/**
 *  Delete catalog recovery policy, job and trigger in case of failure
 *
 * @param step
 * @param funcStream
 * @param ecxSessionId
 * @param oFromNg
 * @param cb
 */
function catalogRecoveryRollback(step, funcStream, ecxSessionId, oFromNg, cb)
{
    ngp.addDoubleSpace();
    ngp.logMsg('info', '    Starting catalog recovery rollback ==>');
    ngp.addDoubleSpace()

	var rollbackState = ngp.getRequestState(step.reqId, ecxSessionId, 'rollbackState');
	var rollbackResult = {};

    ngp.ngpCoreHelper.deleteJob(ecxSessionId, rollbackState['jobId'], function(deleteJobCb){
		rollbackState['jobRollbackAttempted'] = true;

		if ( deleteJobCb.e || ( deleteJobCb.response.statusCode != 204 && deleteJobCb.response.statusCode != 404 && deleteJobCb.response.statusCode != 500 ) ){
			ngp.logMsg('error', '    Rollback deletion of catalog recovery job ' + rollbackState['policyName'] + ' failed because of error = ' + JSON.stringify(deleteJobCb.rO));
			rollbackResult['jobId'] = rollbackState['jobId'];
		}else{
            rollbackState['jobRollbackSuccessful'] = true;

            ngp.ngpCoreHelper.deletePolicy(ecxSessionId, rollbackState['policyId'], function(deletePolicyCb){
                rollbackState['policyRollbackAttempted'] = true;

                if ( deletePolicyCb.e || ( deletePolicyCb.response.statusCode != 204 && deletePolicyCb.response.statusCode != 404 && deletePolicyCb.response.statusCode != 500 ) ){
                    ngp.logMsg('error', '    Rollback deletion of policy ' + rollbackState.policyName + ' failed because of error = ' + JSON.stringify(deletePolicyCb.rO));
                    rollbackResult['policyId'] = rollbackState['policyId'];
                }else{
                    rollbackState['policyRollbackSuccessful'] = true;

                    ngp.ngpCoreHelper.deleteTrigger(ecxSessionId, rollbackState['triggerId'], function(deleteTriggerCb){
                        if ( deleteTriggerCb.e || ( deleteTriggerCb.response.statusCode != 204 && deleteTriggerCb.response.statusCode != 404 && deleteTriggerCb.response.statusCode != 500 ) ){
                            ngp.logMsg('warn', '    Rollback deletion of trigger for policy ' + rollbackState['policyName'] + ' failed because of error = ' + JSON.stringify(deleteTriggerCb.rO));
                            rollbackResult['triggerId'] = rollbackState['triggerId'];
                        }else{
                            rollbackState['triggerRollbackSuccessful'] = true;
                        }

                        var result = {
                            statusCode: 500,
                            response: {
                                failedRollbacks: rollbackResult,
                            }
                        };

                        ngp.logMsg('error', 'Catalog Recovery Rollback State = ' + JSON.stringify(rollbackState));
                        ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, result, false);
                    });
                }
            });
        }
    });
}

exports.catalogRestoreFuncSteps = catalogRestoreFuncSteps;
exports.catalogRestoreFuncs = catalogRestoreFuncs;