if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('../ecxngp');

//Add list fo ecxngp queryr parameters here
//This list will be excluded when calls are passed to core service
var ecxNgpQueryParameters = ['userId'];

/**
 * Converts ECX url to ECXNGP url format
 * 
 * Note: replace incoming strings and port
 * 
 * @param url 
 * @param substr 
 * @param newSubStr 
 */
function convertCoreServiceUrlToEcxngpUrl(url, substr, newSubStr){
    var newStr = url.replace(substr,newSubStr);

    return newStr.replace(ngp.ecxInternal,ngp.SECUREPORT);
}

/**
 * Returns a given request query object as string
 * 
 * @param queryParameterArray 
 */
function getQueryString(requestQueryObj, joinType){
    var queryString = null;
    if (joinType == 'begin'){
        queryString = '?';
    }else{
        queryString = '&';
    }

    for (var prop in requestQueryObj) {
        if (prop == 'filter'){
            queryString += prop + "=" + ngp.escape(JSON.stringify(requestQueryObj[prop])) + "&";    
            queryString = queryString.replace(/%2520/g,'%20');   //This is mainly to handle the space character in filter parameter values            
        }else{
            queryString += prop + "=" + requestQueryObj[prop] + "&";
        }        
    }    
    return queryString.slice(0,-1);
}

/**
 * Formats the incoming request query object 
 * 
 * @param requestQueryObject 
 */
function formatRequestQueryObject(requestQueryObject){
    var newObj = Object.assign({},requestQueryObject);
    var jsonString = JSON.stringify(newObj);    
    var newStr = jsonString.replace(/\\/g,"")
                           .replace("\"[","[")
                           .replace("]\"","]");
    return JSON.parse(newStr); 
}

/**
 * Delete Ecxngp specific query parameters from request query object from
 * calling the core service
 * 
 * @param queryParameterObj 
 */
function removeEcxngpQueryParameters(queryParameterObj){
    var newObj = Object.assign({},queryParameterObj);
    ecxNgpQueryParameters.forEach(function(item, index, array) {
        delete newObj[item];
    });
    return newObj;
}

/**
 * Delete Non-Ecxngp specific query parameters from request query object
 * 
 * @param queryParameterObj 
 */
function getEcxngpQueryParameters(queryParameterObj){
    var newObj = Object.assign({},queryParameterObj);
    Object.keys(newObj).forEach(function(qParameter) {
        if (!ecxNgpQueryParameters.includes(qParameter)){
            delete newObj[qParameter];
        }        
    });
    return newObj;
}

/**
 * Checks if a nested property exists within an object
 * 
 * @param obj Object to check
 * @param key Property key to check
 */
function checkNestedPropertyExists(obj, key) {
    return key.split(".").every(function(x) {
        if(typeof obj != "object" || obj === null || ! x in obj)
            return false;
        obj = obj[x];
        return true;
    });
}

/**
 * Checks if an object is empty
 * 
 * @param obj 
 */
function isObjectEmpty(obj) {
    for(var key in obj) {
        if(obj.hasOwnProperty(key))
            return false;
    }
    return true;
}

exports.convertCoreServiceUrlToEcxngpUrl = convertCoreServiceUrlToEcxngpUrl;
exports.getQueryString = getQueryString;
exports.formatRequestQueryObject = formatRequestQueryObject;
exports.removeEcxngpQueryParameters = removeEcxngpQueryParameters;
exports.getEcxngpQueryParameters = getEcxngpQueryParameters;
exports.checkNestedPropertyExists = checkNestedPropertyExists;
exports.isObjectEmpty = isObjectEmpty;