if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');   

/*
  Set 'skip' in create_policy if ALL of the subtype X slapolicies already exists. Could be multiple policies/jobs to create. I think you should push all of these from is_policy to 
  create_policy.
*/

var funcSteps = [                                                                                 
	{                                                                                              
		continue: 'next',                                                                          
		name: 'crp_ispolicy',
		verb: 'GET',
		path: '/api/endeavour/policy',
		ip: null,                                                                                                                                      
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                            // set by previous step [ next | skip | fail ]     
		name: 'crp_create',   
		verb: 'POST',
		path: '/api/endeavour/policy',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                       
		name: 'crp_link',   
		verb: 'POST',
		path: '/api/application?from=HLO&action=applyStorageProfiles',       		                                    
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
];


// NB: rollback logic herein could be a candidate for failsafe logic as in regprovider.js

const crProtPolStepFuncs = {
	crp_ispolicy: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		ngp.isProtectionPolicy(step, funcStream, ecxSessionId, oFromNg, cb);
	},
	crp_create: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		createPolicy(step, funcStream, ecxSessionId, oFromNg, cb);      		                               
	},
	crp_link: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		linkResources(step, funcStream, ecxSessionId, oFromNg, cb);      		                               
	},	
};


exports.init = init;


function init(req, res, providerType)                                    
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                            // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                     // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.func = crProtPolStepFuncs[step.name];                                                  // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'create_' + 'application_protection_policy');  

	ngState.currentFunc = 'create_' + 'application_protection_policy';                                 
	return stepsStream;
}


var linkResources = function(step, funcStream, ecxSessionId, oFromNg, cb)                          // finally drive ECX '/api/applicationcal?action=applyStorageProfiles', using NG POST body	
{
	var slapArClone = JSON.parse(JSON.stringify(oFromNg.slapolicies));                               // need to add this to oFromNg under property name storageProfiles
	delete oFromNg.slapolicies;
	oFromNg.storageProfiles = slapArClone;
	
	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, oFromNg, function(lrO){   
		if ( lrO.e || ((lrO.response.statusCode != 200) && (lrO.response.statusCode != 201)))            		                                 //!! REPLACE with 'if ( true )' to force rollback in this step
		{
			var finishedPjEntries = ngp.getRequestState(step.reqId, ecxSessionId, 'finishedPjEntries');  // stuff what may need rolling back

			if ( ! finishedPjEntries || finishedPjEntries.length < 1 )                                   // EHI likely that policy create step was skipped, never setting finishedPjEntries
			{
					var failResult = ngp.getGenericEcxErrorResponse(lrO, ecxSessionId, step);
					ngp.logMsg('warn', 'Detected failure of policy and job resource linkage in session ' + ecxSessionId + '; NOTHING TO ROLLBACK'); 
					ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
					return;
			}
			
			ngp.logMsg('warn', 'Detected failure of policy and job resource linkage in session ' + ecxSessionId + '; attempting rollback'); 

			var rolledbackPolJobsAr = [];

			rollbackPolicyAndJob(finishedPjEntries, ecxSessionId, rolledbackPolJobsAr, function(){    
				var failResult = {
					statusCode: 500,                                                             
					response: {
						rollbackResults: rolledbackPolJobsAr,
					},                                                       
				};

				ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
			});
		}
		else
		{
			var crSummaryResults = ngp.getRequestState(step.reqId, ecxSessionId, 'policiesCreationResults');    // push to next step in case needed (successful) !!hypp

			var successResult = {
				statusCode: lrO.response.statusCode,
				response: crSummaryResults,    
			};
			
			ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
		}
	});
};


var createPolicy = function(step, funcStream, ecxSessionId, oFromNg, cb)                           
{
	//@#@ var createPolNamesHash = ngp.getEndpointState(ecxSessionId, 'createPolicyNames');
	var createPolNamesHash = ngp.getRequestState(step.reqId, ecxSessionId, 'createPolicyNames');
	
	//ngp.logMsg('warn', 'POLNAMESHASH=' + JSON.stringify(createPolNamesHash));

	var slapHash = ngp.arrayToHash(oFromNg.slapolicies, 'name');                                     // Input from client
	var pjEntries = [];                                                                                                                                         
	
	var rollback = {
		policies: [],
		jobs: [],
	};

	var policyNow = (new Date).getTime();

	for ( policyName in createPolNamesHash )
	{
		var pjEntry = {};
		var temp = policyName.split('_');                                                           // type, e.g., vmware, in element 0
		var entityType = temp.shift();
		var slapMetal = temp.join('_');

		//ngp.logMsg('warn', 'METAL=' + slapMetal);
		var slap = slapHash[slapMetal];

		var apppolTplClone = JSON.parse(JSON.stringify(ngp.applicationProtectionPolicyTemplate));
		apppolTplClone.name = policyName;
		apppolTplClone.subType = entityType;
		apppolTplClone.spec.storageworkflow[0].id = slap.id;
		apppolTplClone.spec.storageworkflow[0].name = slap.name;
		apppolTplClone.spec.storageworkflow[0].href = slap.href;

		//!!hyppolTplClone.spec.storageworkflow[0].trigger.activateDate = policyNow;                 // ECC-10979
		delete apppolTplClone.spec.storageworkflow[0].trigger;
		
		var pjEntry = {
			policyName: policyName,
			policyCreated: false,
			policyId: null,
			jobCreated: false,
			jobId: null,
			policyPostBody: apppolTplClone,                                                          // created up front, outbound to ECX via POST
			policyRollbackAttempted: false,
			policyRollbackSuccessful: false,
			jobRollbackAttempted: false,
			jobRollbackSuccessful: false,	                                                         // NB: set on 204 or 404 from DELETE
		};
		
		pjEntries.push(pjEntry);                                                                     // unlike recovery case, here we can have multiple policy/job pairs to create
	}

	//ngp.logMsg('warn', 'HMR=' + JSON.stringify(pjEntries));
	var numPoliciesToCreate = pjEntries.length;

	var finishedPjEntries = [];                                                                      // 'finished' as in done with; maybe entity created, maybe not
	var crSummaryResults = [];                                                                       // will contain results of successful policy/job creation in pairs


// At any point in time pjEntries will contain unfinished (not yet presented to ECX pjEntries each of which represents a policy/job pair) and finishedPjEntries will contain finished. 
// The entities represented by an element in finishedPjEntries may or may not have been created. Stuff remaining in pjentries never got created and thus can't be rolled back.

	createPolicyAndJob(pjEntries, step, ecxSessionId, finishedPjEntries, crSummaryResults, function(failResult){
		if ( failResult )                                                                            // FTI rollback anything in finishedPjEntries that indicates it was created
		//if ( true )                        // uncomment to force rollback               
		{
			ngp.logMsg('warn', 'Detected failure of policy and job creation in session ' + ecxSessionId + '; attempting rollback'); 

			var rolledbackPolJobsAr = [];
			
			rollbackPolicyAndJob(finishedPjEntries, ecxSessionId, rolledbackPolJobsAr, function(){    
				var failedResult = {
					statusCode: failResult.statusCode,
					response: {rc: failResult.response.rc}																
				};
								
				ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failedResult, false);    // we're done; link_resources got nothing to link
			});
		}
		else
		{
			var rollbackState = {
				createdPolJobsAr: finishedPjEntries,
				rolledbackPolJobsAr: rolledbackPolJobsAr,                                            // !! perforce empty cuz all were created OK, but make explicit
			};

			ngp.setRequestState(step.reqId, ecxSessionId, 'policiesCreationResults', crSummaryResults);         // push to next step in case needed (successful) 
			ngp.setRequestState(step.reqId, ecxSessionId, 'finishedPjEntries', finishedPjEntries);              // push to next step in case it fails
			
			step.cb(ecxSessionId, null);                                                             // nothing to return here: many ops may have been performed (at least 2)                                                                                       
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                               // drive next step only after all policies & jobs have been created
		}		
	});
}


// NB: no trigger logic required for protection policies; handled by backend

// createPolicyAndJob only returns after all entries in pjEntries (each entry represents a policy/job pair that must be created) have been processed; or after a failure

function createPolicyAndJob(pjEntries, driverStep, ecxSessionId, finishedPjEntries, crSummaryResults, policiesCreatedCb)    // RECURSIVE function
{
	if ( pjEntries.length < 1 )
	{
		policiesCreatedCb();
	}
	else
	{
		var pjEntry = pjEntries.shift();                                                             // leftmost element in array of pjEntry objects
		
		finishedPjEntries.push(pjEntry);                                                             // flags set (or not) below will determine need for rollback, i.e., something created but then some failure

		ngp.sendRequestToEcx(driverStep.verb, ngp.ecxExternal, driverStep.path, ecxSessionId, pjEntry.policyPostBody, function(appObj){    // Create POLICY
			if ( appObj.e || appObj.response.statusCode != 201 )                                 // failed to create the ECX policy  
			{
				var failResult = {
					statusCode: appObj.response.statusCode,
					response: {rc: (JSON.stringify(appObj.rO.description).replace(/\"/g, ""))}																
				};
				
				policiesCreatedCb(failResult);			                                             // return to rollback first jobs, then policies
			}   
			else                                                                                     // a single policy was created, but non-terminal cuz we still need to drive link_resources
			{
				crSummaryResults.push(appObj.rO);                                                  // save to give back to caller; an array of the created policy/job pairs as returned by ECX

				pjEntry.policyId = appObj.rO.id;
				pjEntry.policyCreated = true;
				
				ngp.logMsg('debug', "Application protection policy creation response=" + JSON.stringify(appObj.rO)); 

				var jobTplClone = JSON.parse(JSON.stringify(ngp.jobTpl));

				jobTplClone.name = pjEntry.policyName;
				jobTplClone.description += pjEntry.policyName;
				jobTplClone.policyId = appObj.rO.id;

				ngp.sendRequestToEcx(driverStep.verb, ngp.ecxExternal, '/api/endeavour/job', ecxSessionId, jobTplClone, function(hyppjObj){    // create JOB
					if ( hyppjObj.e || hyppjObj.response.statusCode != 201 )
					{
						var failResult = {
							statusCode: hyppjObj.response.statusCode,
							response: {rc: (JSON.stringify(hyppjObj.rO.description).replace(/\"/g, ""))}																
						};
						
						policiesCreatedCb(failResult);
					}
					else
					{
						crSummaryResults.push(hyppjObj.rO);                                          // save to give back to caller 

						pjEntry.jobId = hyppjObj.rO.id;
						pjEntry.jobCreated = true;
						
						ngp.logMsg('debug', "Created application protection job=" + JSON.stringify(hyppjObj.rO)); 

						createPolicyAndJob(pjEntries, driverStep, ecxSessionId, finishedPjEntries, crSummaryResults, policiesCreatedCb);    // RECURSE
					}
				});
			}
		});		
	}
}


// Attempt to rollback all pjEntries (each represents a policy/job pair) regardless of whether or not they show successful creation. If entity ID is null because, then you will likely get a 500 from ECX.

function rollbackPolicyAndJob(finishedPjEntries, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb)              
{
	if ( finishedPjEntries.length < 1 )
	{
		policiesRolledbackCb();
	}
	else
	{
		var pjEntry = finishedPjEntries.shift();

		if ( ! pjEntry.policyCreated )                                                                                                      // NB: if policy not created, then neither was its job
		{
			ngp.logMsg('warn', "Policy and job " + pjEntry.policyName + " were never created; nothing to rollback");  
			rollbackPolicyAndJob(finishedPjEntries, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb);                               // RECURSE
		}
		else
		{
// NB: to handle policy but no job you prob have to put an 'if' block here that tests for ! pjEntry.jobCreated. If it wasn't created then drive SRTE to delete the policy and then recurse.
// else fall thru to drive both job & policy deletion. It'd be cleaner if the normative deletion of policy after job weren't in callback after successful job deletion.
			
			ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, '/api/endeavour/job/' + pjEntry.jobId, ecxSessionId, null, function(jrObj){    // !! DELETE JOB, then its POLICY
				pjEntry.jobRollbackAttempted = true;
				
				if ( jrObj.e || ( jrObj.response.statusCode != 204 && jrObj.response.statusCode != 404 && jrObj.response.statusCode != 500 ) )
				{
					ngp.logMsg('warn', "Rollback deletion of job " + pjEntry.policyName + " failed because of error=" + JSON.stringify(jrObj.rO));  
				}
				else
					pjEntry.jobRollbackSuccessful = true;

	// FTI job may or may not have been deleted. Now nuke its policy.

				ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, '/api/endeavour/policy/' + pjEntry.policyId, ecxSessionId, null, function(prObj){    // delete policy
					pjEntry.policyRollbackAttempted = true;				
					
					if ( prObj.e || ( prObj.response.statusCode != 204 && prObj.response.statusCode != 404 && prObj.response.statusCode != 500 ) )
					{
						ngp.logMsg('warn', "Rollback deletion of policy " + pjEntry.policyName + " failed because of error=" + JSON.stringify(jrObj.rO));  
						rolledbackPolJobsAr.push(pjEntry);
						rollbackPolicyAndJob(finishedPjEntries, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb);       // RECURSE
					}
					else
					{
						pjEntry.policyRollbackSuccessful = true;
						rolledbackPolJobsAr.push(pjEntry);
						rollbackPolicyAndJob(finishedPjEntries, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb);       // RECURSE
					}
				});
			});	
		}	
	}
}
