if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');   

/*
  Set 'skip' in create_policy if ALL of the subtype X slapolicies already exists. Could be multiple policies/jobs to create. I think you should push all of these from is_policy to 
  create_policy.
*/

var funcSteps = [                                                                                  // generalized user handling + provider registration
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // first constituent always continues
		name: 'crp_ispolicy',
		verb: 'GET',
		path: '/api/endeavour/policy',
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'crp_create',   
		verb: 'POST',
		path: '/api/endeavour/policy',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: nothing in this block will change
		continue: 'next',                                                       
		name: 'crp_link',   
		verb: 'POST',
		path: '/api/storage?action=applyStorageProfiles',       		                                      
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
];


const crProtPolStepFuncs = {
	crp_ispolicy: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		ngp.isProtectionPolicy(step, funcStream, ecxSessionId, oFromNg, cb);
	},
	crp_create: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		createPolicy(step, funcStream, ecxSessionId, oFromNg, cb);      		                               
	},
	crp_link: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		linkResources(step, funcStream, ecxSessionId, oFromNg, cb);      		                               
	},	
};


exports.init = init;


function init(req, res, providerType)                                    
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                            // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                     // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.func = crProtPolStepFuncs[step.name];                                                  // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'create_' + 'storage_protection_policy');  

	ngState.currentFunc = 'create_' + 'storage_protection_policy';                                 
	return stepsStream;
}


var linkResources = function(step, funcStream, ecxSessionId, oFromNg, cb)                            // finally drive ECX '/api/hypervisor?action=applyStorageProfiles', using NG POST body	
{
    delete oFromNg.subtype;                                                                          // except for this property

	var slapArClone = JSON.parse(JSON.stringify(oFromNg.slapolicies));                               // need to add this to oFromNg under property name storageProfiles
	delete oFromNg.slapolicies;
	oFromNg.storageProfiles = slapArClone;
	
	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, oFromNg, function(lrO){   
		if ( lrO.e || lrO.response.statusCode != 201 )             
		{
			ngp.logMsg('warn', 'Detected failure of policy and job resource linkage in session ' + ecxSessionId + '; attempting rollback'); 

			var failResult = ngp.getGenericEcxErrorResponse(lrO, ecxSessionId, step);
			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);
			
			//@#@ var rollbackState = ngp.getEndpointState(ecxSessionId, 'rollbackState');
			var rollbackState = ngp.getRequest(step.reqId, ecxSessionId, 'rollbackState');

			if ( rollbackState )                                                                     // NB: if we skipped creation step because all policies & jobs pre-exist, then no rollback info
			{
				if ( ! rollbackState.rolledbackPolJobsAr )                                           // again, silly, cuz if we in this step all were created; nothing rolledback
					rollbackState.rolledbackPolJobsAr = [];

				rollbackPolicyAndJob(rollbackState.createdPolJobsAr, ecxSessionId, rollbackState.rolledbackPolJobsAr, function(){    
					var failResult = {
						statusCode: 500,                                                             //!! ??
						response: {
							failedRollbacks: rollbackState.rolledbackPolJobsAr,
						},                                                       
					};
				});
			}
		}
		else                                                                                         // FTI everything worked. Return every policy and job object that we created
		{
			//@#@ var crSummaryResults = ngp.getEndpointState(ecxSessionId, 'policiesCreationResults');    // push to next step in case needed (successful) !!hypp
			var crSummaryResults = ngp.getRequestState(step.reqId, ecxSessionId, 'policiesCreationResults');    // push to next step in case needed (successful) !!hypp

			var successResult = {
				statusCode: lrO.response.statusCode,
				//response: lrO.rO,
				response: crSummaryResults,    
			};
			
			ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
		}
	});
};


var createPolicy = function(step, funcStream, ecxSessionId, oFromNg, cb)                           
{
	//@#@ var createPolNamesHash = ngp.getEndpointState(ecxSessionId, 'createPolicyNames');
	var createPolNamesHash = ngp.getRequestState(step.reqId, ecxSessionId, 'createPolicyNames');

	//ngp.logMsg('warn', 'POLNAMESHASH=' + JSON.stringify(createPolNamesHash));

	var slapHash = ngp.arrayToHash(oFromNg.slapolicies, 'name');                                     // Input from client
	var crPolicySteps = [];                                                                                                                                         
	
	var countdownLatch = Object.keys(createPolNamesHash).length;                                     // must create this many policies and their jobs !! VESTIGIAL
	
	var rollback = {
		policies: [],
		jobs: [],
	};

	for ( policyName in createPolNamesHash )
	{
		var pjEntry = {};
		var temp = policyName.split('_');                                                           // type, e.g., vmware, in element 0
		var entityType = temp.shift();
		var slapMetal = temp.join('_');

		//ngp.logMsg('warn', 'METAL=' + slapMetal);
		var slap = slapHash[slapMetal];

		var stppolTplClone = JSON.parse(JSON.stringify(ngp.stppolTpl));
		stppolTplClone.name = policyName;
		//!! BUSTED: stppolTplClone.subtype = entityType;
		stppolTplClone.subType = entityType;
		//ngp.logMsg('warn', 'SUBTYPE=' + stppolTplClone.subType);
		stppolTplClone.spec.storageworkflow[0].id = slap.id;
		stppolTplClone.spec.storageworkflow[0].name = slap.name;
		stppolTplClone.spec.storageworkflow[0].href = slap.href;
		
		var pjEntry = {
			policyName: policyName,
			policyCreated: false,
			jobCreated: false,
			policyPostBody: stppolTplClone,
			policyRollbackAttempted: false,
			policyRollbackSuccessful: false,
			jobRollbackAttempted: false,
			jobRollbackSuccessful: false,	                                                         // NB: set on 204 or 404 from DELETE
		};

		crPolicySteps.push(pjEntry);
	}

	//ngp.logMsg('warn', 'HMR=' + JSON.stringify(crPolicySteps));
	var numPoliciesToCreate = crPolicySteps.length;

	var createdPolJobsAr = [];                                                                       // maybe rollback everything in this array, JOBS FIRST, then their Policies
	var crSummaryResults = [];                                                                       // will contain results of successful policy/job creation in pairs
	
	createPolicyAndJob(crPolicySteps, step, ecxSessionId, createdPolJobsAr, crSummaryResults, function(failResult){    //!! ARE U USING failResult here?  
		if ( createdPolJobsAr.length != numPoliciesToCreate )                                        // NB: FTI not all policies got created; this is failure case, but 1st rollback
		//!! swap with preceding to test rollback:     if ( createdPolJobsAr.length != 0 )                                      
		{
			ngp.logMsg('warn', 'Detected failure of policy and job creation in session ' + ecxSessionId + '; attempting rollback'); 

			var rolledbackPolJobsAr = [];
			
			rollbackPolicyAndJob(createdPolJobsAr, ecxSessionId, rolledbackPolJobsAr, function(){    
				var failResult = {
					statusCode: 500,                                                                 //!! ??
					response: {
						failedRollbacks: rolledbackPolJobsAr,                                        // return what wasn't rolled back
					},                                                       
				};
				ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    // we're done; link_resources got nothing to link
			});
		}
		else                                                                                         // yes, policy & job were created but applyStorageProfiles might fail; need to rollback in that case, too
		{
			var rollbackState = {
				createdPolJobsAr: createdPolJobsAr,
				rolledbackPolJobsAr: rolledbackPolJobsAr,                                            // !! perforce empty cuz all were created OK, but make explicit
			};

			//@#@ ngp.setEndpointState(ecxSessionId, 'policiesCreationResults', crSummaryResults);         // push to next step in case needed (successful)
			//@#@ ngp.setEndpointState(ecxSessionId, 'rollbackState', rollbackState);                      // push to next step in case needed (failure)

			ngp.setRequestState(step.reqId, ecxSessionId, 'policiesCreationResults', crSummaryResults);         // push to next step in case needed (successful)
			ngp.setRequestState(step.reqId, ecxSessionId, 'rollbackState', rollbackState);                      // push to next step in case needed (failure)

			step.cb(ecxSessionId, null);                                                             // nothing to return here: many ops may have been performed (at least 2)                                                                                       
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                               // drive next step only after all policies & jobs have been created
		}
	});
}


function createPolicyAndJob(crPolicySteps, driverStep, ecxSessionId, createdPolJobsAr, crSummaryResults, policiesCreatedCb)    // recursive func; driverStep belongs to the runSerialFuncs in ecxngp
{
	if ( crPolicySteps.length < 1 )
	{
		policiesCreatedCb();
	}
	else
	{
		var crPolicyStep = crPolicySteps.shift();

		ngp.sendRequestToEcx(driverStep.verb, ngp.ecxExternal, driverStep.path, ecxSessionId, crPolicyStep.policyPostBody, function(hypprObj){    // Create POLICY 
			if ( hypprObj.e || hypprObj.response.statusCode != 201 )                                 // failed to create the ECX policy  
			{
				var failResult = ngp.getGenericEcxErrorResponse(hypprObj, ecxSessionId, driverStep);
				policiesCreatedCb(failResult);			                                             // return to rollback first jobs, then policies
			}   
			else                                                                                     // a single policy was created, but non-terminal cuz we still need to create its job & finally drive link_resources
			{
				crSummaryResults.push(hypprObj.rO);                                                  // save to give back to caller
                //ngp.logMsg('warn', "POL RES=" + JSON.stringify(hypprObj.rO));
				crPolicyStep.policyId = hypprObj.rO.id;
				crPolicyStep.policyCreated = true;

				ngp.logMsg('warn', "Storage protection POLICY creation response=" + JSON.stringify(hypprObj.rO));  
				var jobTplClone = JSON.parse(JSON.stringify(ngp.jobTpl));

				jobTplClone.name = crPolicyStep.policyName;
				jobTplClone.description += crPolicyStep.policyName;
				jobTplClone.policyId = hypprObj.rO.id;

				ngp.sendRequestToEcx(driverStep.verb, ngp.ecxExternal, '/api/endeavour/job', ecxSessionId, jobTplClone, function(hyppjObj){    // create JOB
					if ( hyppjObj.e || hyppjObj.response.statusCode != 201 )
					{
						var failResult = ngp.getGenericEcxErrorResponse(hyppjObj, ecxSessionId, driverStep);
						policiesCreatedCb(failResult);
					}
					else
					{
						crSummaryResults.push(hyppjObj.rO);                                             // save to give back to caller 
                        //ngp.logMsg('warn', "JOB RES=" + JSON.stringify(hyppjObj.rO));
						crPolicyStep.jobId = hyppjObj.rO.id;
						crPolicyStep.jobCreated = true;
						createdPolJobsAr.push(crPolicyStep);						                 // make available to caller, eventually
						ngp.logMsg('debug', "Storage protection JOB creation response=" + JSON.stringify(hyppjObj.rO));  
						
						createPolicyAndJob(crPolicySteps, driverStep, ecxSessionId, createdPolJobsAr, crSummaryResults, policiesCreatedCb);    // RECURSE 
					}
				});
			}
		});		
	}
}


function rollbackPolicyAndJob(crRollbackSteps, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb)              
{
	if ( crRollbackSteps.length < 1 )
	{
		policiesRolledbackCb();
	}
	else
	{
		var crRollbackStep = crRollbackSteps.shift();

// NB: to simplify this logic, we always attempt to delete the job, then the policy. If the job wasn't created, we'll simply get a 404. 
		
		ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, '/api/endeavour/job/' + crRollbackStep.jobId, ecxSessionId, null, function(jrObj){    // !! DELETE JOB, then its POLICY
			crRollbackStep.jobRollbackAttempted = true;
			
			if ( jrObj.e || ( jrObj.response.statusCode != 204 && jrObj.response.statusCode != 404 ) )    // NB: tolerate, perhaps unwisely, a not found
			{
				ngp.logMsg('warn', "Deletion of job " + crRollbackStep.policyName + " failed because of error=" + JSON.stringify(jrObj.rO));  
			}
			else
				crRollbackStep.jobRollbackSuccessful = true;

// FTI job may or may not have been deleted. Now nuke its policy.

			ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, '/api/endeavour/policy/' + crRollbackStep.policyId, ecxSessionId, null, function(prObj){    // delete policy
				crRollbackStep.policyRollbackAttempted = true;				
				
				if ( prObj.e || ( prObj.response.statusCode != 204 && prObj.response.statusCode != 404 ) ) 
				{
					ngp.logMsg('warn', "Deletion of policy " + crRollbackStep.policyName + " failed because of error=" + JSON.stringify(jrObj.rO));  
					rolledbackPolJobsAr.push(crRollbackStep);
					rollbackPolicyAndJob(crRollbackSteps, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb);       // RECURSE
				}
				else
				{
					crRollbackStep.policyRollbackSuccessful = true;
					rolledbackPolJobsAr.push(crRollbackStep);
					rollbackPolicyAndJob(crRollbackSteps, ecxSessionId, rolledbackPolJobsAr, policiesRolledbackCb);       // RECURSE
				}
			});
		});		
	}
}
