if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

    
var ngp = require('./ecxngp');   

var funcStepsPost = [                                                                                  
    {                                                                                              
      continue: 'next',                                                                              
      name: 'cru_createrpool',                                                                    
      verb: 'POST',
      path: '/api/security/resourcepool',                                                            
      ip: null, 
      port: null, 
      type: null,
      cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },     
    {                                                                                              
      continue: 'next',                                                                            
      name: 'cru_createuser',   
      verb: 'POST',
      path: '/api/security/user',                                                            
      ip: null, 
      port: null, 
      type: null,
      cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
    {                                                                                              
      continue: 'next',                                                                                 // driven only if previous step fails                                                                     
      name: 'cru_deleterpool',   
      verb: 'DELETE',
      path: '/api/security/resourcepool/',                                                              // we'll append resourcepool id here in case we need to drive rollback                                                   
      ip: null, 
      port: null, 
      type: null,
      cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },    
];

var funcStepsPut = [                                                                                  
    {                                                                                              
      continue: 'next',                                                                              
      name: 'cru_updateuser',                                                                    
      verb: 'PUT',
      path: '/api/security/user/',                                                                      // !!! THINK WE NEED ../USER_ID?resourcePoolId=RPOOL_ID                                                        
      ip: null, 
      port: null, 
      type: null,
      cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },     
];

var funcStepsDelete = [                                                                                  
    {                                                                                              
      continue: 'next',                                                                              
      name: 'cru_del_deleteuser',                                                                    
      verb: 'DELETE',
      path: '/api/security/user/',                                                            
      ip: null, 
      port: null, 
      type: null,
      cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },     
    {                                                                                              
      continue: 'next',                                                                            
      name: 'cru_del_deleterpool',   
      verb: 'DELETE',
      path: '/api/security/resourcepool/',                                                            
      ip: null, 
      port: null, 
      type: null,
      cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
    },
];

const cruserStepFuncs = {
    cru_createrpool: function(step, funcStream, ecxSessionId, oFromNg, cb){                               
      createResourcePool(step, funcStream, ecxSessionId, oFromNg, cb);   
    },
    cru_createuser: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   
      createUser(step, funcStream, ecxSessionId, oFromNg, cb);    
    },
    cru_deleterpool: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   
      deleteResourcePool(step, funcStream, ecxSessionId, oFromNg, cb);    
    }, 
    cru_del_deleteuser: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   
      deleteUser(step, funcStream, ecxSessionId, oFromNg, cb);    
    }, 
    cru_del_deleterpool: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   
      deleteDeleteResourcePool(step, funcStream, ecxSessionId, oFromNg, cb);                
    },              	
    cru_updateuser: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   
      updateUser(step, funcStream, ecxSessionId, oFromNg, cb);    
    },
};


exports.init = init;


function init(req, res, subType)                                                                       
{
    var ngState = ngp.getSessionState(req, res);
    
    if ( ngState.stale )                                                                                 
      return;                                                                                            

    var funcSteps = null;
    var genericFuncName = null;

    if ( req.method == 'PUT' )
    {
        genericFuncName = 'update_userrpool';
        funcSteps = JSON.parse(JSON.stringify(funcStepsPut));
    }   

    if ( req.method == 'POST' )
    {
        genericFuncName = 'create_userrpool';
        funcSteps = JSON.parse(JSON.stringify(funcStepsPost));
    }   
    
    if ( req.method == 'DELETE' )
    {
        genericFuncName = 'delete_userrpool';
        funcSteps = JSON.parse(JSON.stringify(funcStepsDelete));
    } 

    var ecxSid = ngState.ecxSid;
    res.ecxSid = ecxSid;

    var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

    stepsStream.forEach(function(step){
      step.ip = ngp.localIp;
      step.port = ngp.ecxExternal;

      step.func = cruserStepFuncs[step.name];     
      
      step.cb = function(ecxSid, data){ 
        ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

        if ( data )
        {
          ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
        }
      };
    });

    ngp.createRequestStore(stepsStream, ecxSid, req.reqId, genericFuncName);                            // add reqId to steps, create store for the req & set currentFunc to genericFuncName
    ngp.setRequestState(req.reqId, ecxSid, 'crudelUrl', req.url);                                       // used by delete_userrpool steps
    ngp.setRequestState(req.reqId, ecxSid, 'cruupdUrl', req.url);                                       // used by update_userrpool steps; same as what we just stored (keep it clean)

    ngState.currentFunc = genericFuncName;                                                               
    return stepsStream;
}


function createResourcePool(step, funcStream, ecxSessionId, oFromNg, cb)
{
    var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);
    ngp.logMsg('warn', "Input from client for create user in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

    if ( oFromNg.roleIds.length == 1 && oFromNg.roleIds[0] == '1001' )                                  // FTI administrator case (vide ECC-12142)
    {
        ngp.setRequestState(step.reqId, ecxSessionId, 'cruserModeAdmin', true);  
        
        ngp.logMsg('warn', "Detected administrator case, SKIPPING CREATION OF RP");
        
        step.cb(ecxSessionId, null);                                                  
        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                                      // move to next step, i.e., don't create RP, nor store state for it
        return;
    }

// EHI the non-administrator case; create RP as before ECC-12142.

    var postBody = {
        name: oFromNg.name + '_rpool',
        description: '',
        resources: [],
        /*
        resources: [
            {
                path: 'root:0/site:0',
                include: true,
            }
        ],
        */
    };

    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(rpO){ 
        if ( rpO.e || rpO.response.statusCode != 201 )   
        {
          var failResult = ngp.getGenericEcxErrorResponse(rpO, ecxSessionId, step); 
          ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);       				
        }
        else                                                                                            // EHI resourcepool created
        {
          ngp.logMsg('warn', "Resourcepool creation OK response=" + JSON.stringify(rpO.rO));  
          ngp.setRequestState(step.reqId, ecxSessionId, 'createdRpool', rpO.rO);

          step.cb(ecxSessionId, rpO.rO);                                                  
          ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 					
        }
    });    
}


function createUser(step, funcStream, ecxSessionId, oFromNg, cb)
{
    var cruserAdminMode = ngp.getRequestState(step.reqId, ecxSessionId, 'cruserModeAdmin');

    var rpO = ngp.getRequestState(step.reqId, ecxSessionId, 'createdRpool');                            // NB: null if this is non-admin case, i.e., we didn't create the rpool

    var effRpId = null;                                                                                 

    if ( cruserAdminMode )                                                                              // FTI bad assumption
    {
        effRpId = "1001";
    }
    else                                                                                                // EHI we created RP and thus rpO is valid
        effRpId = rpO.id;                                                                               // assume non-administrator case
    

    var postBody = {
        name: oFromNg.name,
        password: oFromNg.password,
        type: oFromNg.type,
        sourceProvider: oFromNg.sourceProvider,
        sourceUser: oFromNg.sourceUser,
        sourceUserDN: oFromNg.sourceUserDN,
        permissions: [
            {
                resourcePoolId: effRpId,
                roleIds: oFromNg.roleIds,                                                               // NB: in administrator case this will be ["1001"]
            }
        ]
    };

    ngp.logMsg('debug', "Creating user=" + JSON.stringify(postBody));  
    
    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(usrO){ 
        if ( usrO.e || usrO.response.statusCode != 201 )   
        {
            var failResult = ngp.getGenericEcxErrorResponse(usrO, ecxSessionId, step); 
             ngp.setSerialStreamResult(ecxSessionId, failResult);                                        // let rollback run, but THIS error is final status
             
             if ( ! rpO )                                                                               // FTI we didn't create resource pool, so skip its deletion in rollback step
             {
                ngp.logMsg('warn', "CANNOT ROLLBACK DELETE RP WE DIDNT CREATE");  

                if ( funcStream.length > 0 )                                                            // this has become sort of silly, pretending that the code is unaware of the step order 
                    funcStream[0].continue = 'skip';                                                    // thus no need for rollback (deletion of the resource pool)
             }

            step.cb(ecxSessionId, null);                                                  
            ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	   				                // let next step (rollback) run (or not)
        }
        else                                                                                            // EHI user created
        {
            ngp.logMsg('debug', "User creation OK response=" + JSON.stringify(usrO.rO));  

            if ( funcStream.length > 0 )                                                                // this has become sort of silly, pretending that the code is unaware of the step order 
                funcStream[0].continue = 'skip';                                                        // thus no need for rollback (deletion of the resource pool)
            
            var successResult = {
                statusCode: usrO.response.statusCode,
                response: usrO.rO,    
            };
            
          	ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);		
        }
    }); 
}


function deleteResourcePool(step, funcStream, ecxSessionId, oFromNg, cb)                                // NB: this is ROLLBACK for *CREATE USER* case if user create fails
{
    ngp.logMsg('warn', 'Detected failure of user creation in session ' + ecxSessionId + '; attempting rollback'); 

    var rpO = ngp.getRequestState(step.reqId, ecxSessionId, 'createdRpool');                            // by prior logic this must be defined, i.e., ref to an RP object

    step.path += rpO.id;

    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, null, function(delrpO){   
        if ( delrpO.e || ( delrpO.response.statusCode != 204 ) )
        {
            ngp.logMsg('warn', "Rollback deletion of resourcepool " + oFromNg.name + '_rpool' + " failed because of error=" + JSON.stringify(delrpO.rO));  
        }

        step.cb(ecxSessionId, null);                                                  
        ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);
    });           		    
}


function updateUser(step, funcStream, ecxSessionId, oFromNg, cb)
{
    var cruupdUrl = ngp.getRequestState(step.reqId, ecxSessionId, 'cruupdUrl');   
    var cruupdUserId = cruupdUrl.match(/\/(\d+)$/)[1]; 

    var putBody = {
        name: oFromNg.name,
        /* Per ECC-12142
        permissions: [
          {
              resourcePoolId: cruupdRpoolId,
              roleIds: oFromNg.roleIds,
          }
        ]
        */
    };

    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path + cruupdUserId, ecxSessionId, putBody, function(usrO){ 
        if ( usrO.e || usrO.response.statusCode != 200 )   
        {
            var failResult = ngp.getGenericEcxErrorResponse(usrO, ecxSessionId, step); 
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);       				
        }
        else                                                                                            // EHI user updated
        {
            ngp.logMsg('debug', "User update OK response=" + JSON.stringify(usrO.rO));  

            var successResult = {
                statusCode: usrO.response.statusCode,
                response: usrO.rO,    
            };
            
          	ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);		
        }
    }); 
}


// Next two 'delete' functions are part of the DELETE USER life-cycle function initiated via DELETE /ecxngp/security/user/<USER_ID>?resourcePoolId= <RP_ID>

function deleteUser(step, funcStream, ecxSessionId, oFromNg, cb)                                       
{
    var crudelUrl = ngp.getRequestState(step.reqId, ecxSessionId, 'crudelUrl');                                
    var crudelUserId = crudelUrl.match(/\/(\d+)\?/)[1];
    var crudelRpoolId = crudelUrl.match(/resourcePoolId\=(\d+)/)[1];

    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path + crudelUserId, ecxSessionId, null, function(delusrO){   
        if ( delusrO.e || ( delusrO.response.statusCode != 204 ) )
        {
            ngp.logMsg('warn', 'Deletion of user ID=' + crudelUserId + ' failed because of error=' + JSON.stringify(delusrO.rO)); 
            var failResult = ngp.getGenericEcxErrorResponse(delusrO, ecxSessionId);          
            
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);      // don't run next step (delete rp) cuz it won't work if user still linked
        }
        else                                                                                            // EHI user deleted
        {
            if ( crudelRpoolId == '1001' )                                                              // FTI skip next step (delete RP)
            {
                ngp.logMsg('debug', 'Skipping deletion of resource pool with ID=' + crudelRpoolId); 
                
                if ( funcStream.length > 0 )                                                            
                    funcStream[0].continue = 'skip';                                                        
            
                var successResult = {
                    statusCode: delusrO.response.statusCode,
                    response: null,    
                };
            
          	    ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);		
            }
            else
            {
                step.cb(ecxSessionId, null);                                                  
                ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb);                              // if not RP ID 1001, try to delete rpool in next step
            }
        }
    });    
}


function deleteDeleteResourcePool(step, funcStream, ecxSessionId, oFromNg, cb)                                       
{
    var crudelUrl = ngp.getRequestState(step.reqId, ecxSessionId, 'crudelUrl');                                
    var crudelRpoolId = crudelUrl.match(/resourcePoolId\=(\d+)/)[1];

    ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path + crudelRpoolId, ecxSessionId, null, function(delrpO){   
        if ( delrpO.e || ( delrpO.response.statusCode != 204 ) )
        {
            ngp.logMsg('warn', "Deletion of resourcepool " + oFromNg.name + '_rpool' + " failed because of error=" + JSON.stringify(delrpO.rO));  
            var failResult = ngp.getGenericEcxErrorResponse(delrpO, ecxSessionId);          
            
            ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);      
        }
        else
        {
            var successResult = {
                statusCode: delrpO.response.statusCode,
                response: null,    
            };
            
            ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
        }
    });    
}