if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}

var ngp = require('./ecxngp');

var funcSteps = [                                                                                  
	{                                                                                              
		continue: 'next',                                                                          
		name: 'app_getInstances',
		verb: 'POST',
		path: '/api/appserver?action=discover', 
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,                                                                                
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	} 
];

const appGetInsatncesStatusStepFuncs = {
	app_getInstances: function(step, funcStream, ecxSessionId, oFromNg, cb){                                
		getInstancesStatus(step, funcStream, ecxSessionId, oFromNg, cb); 
	},
};

exports.init = init;

function init(req, res)                                    
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                            
	return;                                                                                     

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	stepsStream.forEach(function(step) {
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.func = appGetInsatncesStatusStepFuncs[step.name];                                                   
		
		step.cb = function(ecxSid, data) { 
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data ) {
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'app_getInstancesStatus');  
	ngState.currentFunc = 'app_getInstancesStatus';                                           
	return stepsStream;
}

var getInstancesStatus = function(step, funcStream, ecxSessionId, oFromNg, cb) { 

	var postBody = {  
		user: oFromNg.user,
		properties: oFromNg.properties,                                                  
		provider: oFromNg.provider,
		level: oFromNg.level                                                        
	};
                      
	ngp.logMsg('warn', "Input from client for lightweight discovery in session ID='" + ecxSessionId + "'=" + postBody);

	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(liveDiscoveryObj) {
		if ( liveDiscoveryObj.e || liveDiscoveryObj.response.statusCode != 200 ) {
			var failResult = ngp.getGenericEcxErrorResponse(liveDiscoveryObj, ecxSessionId, step);			
		    ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);	
	    }
	    else {
			var instancesFromCore = oFromNg.instances;
			
			if(oFromNg.provider == 'mongo')
			var instancesFromLiveDisc = liveDiscoveryObj.rO['instances'];
			if(oFromNg.provider == 'oracle')
			var instancesFromLiveDisc = liveDiscoveryObj.rO['databases'];
					
			var instancesStatusAfterMerge = [];
		
			for(var i=0; i<instancesFromCore.length; i++) {
				var staleStatus = true;
				for(var j=0; j<instancesFromLiveDisc.length; j++) {
					if(instancesFromCore[i].name == instancesFromLiveDisc[j].name) {
						staleStatus = false;
					
						instancesStatusAfterMerge.push(instancesFromLiveDisc[j])	
						instancesStatusAfterMerge[instancesStatusAfterMerge.length-1].status = "registered";
						instancesStatusAfterMerge[instancesStatusAfterMerge.length-1].href = instancesFromCore[i].user;
						instancesFromLiveDisc.splice(j, 1);

						break;
					}  
				}
			
				if(staleStatus == true) {
					instancesStatusAfterMerge.push(instancesFromCore[i])
					instancesStatusAfterMerge[instancesStatusAfterMerge.length-1].status = "stale";
					instancesStatusAfterMerge[instancesStatusAfterMerge.length-1].href = instancesFromCore[i].user;
				}
			}
		
			for(var i=0; i<instancesFromLiveDisc.length; i++) {
				instancesStatusAfterMerge.push(instancesFromLiveDisc[i])
				instancesStatusAfterMerge[instancesStatusAfterMerge.length-1].status = "live";
			}
		
			if(oFromNg.provider == 'mongo') {
				var finalResponse = { 
					instances : instancesStatusAfterMerge 
				};
			}
			if(oFromNg.provider == 'oracle') {
				var finalResponse = { 
					databases : instancesStatusAfterMerge 
				};
			}

			var successResult = {
				statusCode: liveDiscoveryObj.response.statusCode,
				response: finalResponse
			};
			
		    ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	
	    }
	});	
}