if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');                                                                

/*
  NodeJS will use the combination of SLA policy name and the subtype to construct the GET api/endeavour/jobsession?filter:[{“property”: “jobName", “value”: “<subtype>_<slapolicy_name”}]
   call and get ALL jobsessions that we need to populate job sessions that we are interested.
  NodeJS will return all other attributes from the jobsession objects into the response here.  That includes the csv and log link so that AngularUI can call to present log file content at the 3rd level, 
   and allow user to download the log file when it is required (not in the current design).
*/

var funcSteps = [                                                                                  // generalized user handling + provider registration
	{                                                                                              
		continue: 'next',                                                                          // first constituent always continues
		name: 'get_jobsessions',
		verb: 'GET',
		path: '/api/endeavour/jobsession?filter=[{"property":"jobName","value":"qqVVqq"}]&sort=[{"property":"start","direction":"DESC"}]',    // init replaces PH with jobname, like 'vmware_Gold'
		ip: null,                                                                                  //!! set by ecxngp. but maybe you don't really need per step ip and ort                                                     
		port: null, 
		type: null,
		func: null,                                                                                // see initialization below 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
];


const getJobSessionStepFuncs = {
	get_jobsessions: function(step, funcStream, ecxSessionId, oFromNg, cb){                            
		getJobSessions(step, funcStream, ecxSessionId, oFromNg, cb);   
	},
};


exports.init = init;


function init(req, res, keyVals)                                                    
{
	var ngState = ngp.getSessionState(req, res);
	
	if ( ngState.stale )                                                                           // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                    // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

        if ( step.name == 'get_jobsessions' )
        {
            var temp = keyVals.subtype + '_' + keyVals.name;
            step.path = step.path.replace('qqVVqq', temp);
        }

		step.func = getJobSessionStepFuncs[step.name];                                             // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'get_jobsessions');    

	ngState.currentFunc = 'get_jobsessions';                                                       // as opposed to, say, update_storageprofile
	return stepsStream;
}


var getJobSessions = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
	var ngState = ngp.ecxSidHash[ecxSessionId];

	//path: '/api/endeavour/jobsession?filter=[{"property":"jobName","value":"qqVVqq"}]&sort=[{"property":"start","direction":"DESC"}]',    // init replaces PH with jobname, like 'vmware_Gold'
	

	var currentStepPath = step.path;
	var currentStepPathSplitForFilterSort = currentStepPath.split('=');

	var currentStepPathFilter = currentStepPathSplitForFilterSort[1];
	currentStepPathFilter = currentStepPathFilter.split('&');
	currentStepPathFilter = currentStepPathFilter[0];
	
	var currentStepPathSort = currentStepPathSplitForFilterSort[2];
	
	var fixedFilterOption = ngp.escape(currentStepPathFilter);
	var fixedSortOption = '&sort=' + ngp.escape(currentStepPathSort);
	
	var fixedStepPath = currentStepPathSplitForFilterSort[0] + '=' + fixedFilterOption + fixedSortOption;

	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(sessO){ 
		if ( sessO.e || sessO.response.statusCode != 200 )  
		{
			var failResult = ngp.getGenericEcxErrorResponse(sessO, ecxSessionId, step); 
			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
		}
		else
		{
			var successResult = {
				statusCode: sessO.response.statusCode,
				response: sessO.rO,
			};
			
			ngp.setSerialStreamResult(ecxSessionId, successResult);                                // NB: let caller see result of successful update

			step.cb(ecxSessionId, sessO.rO);                                                       // From ECX callback now drive the recursive driver to move to next constituent function, if any                                    
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 					
		}
	});
};
