if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');                                                                


var funcStepsGet = [                                                                              
	{                                                                                              
		continue: 'next',                                                                          
		name: 'get_issprof',
		verb: 'GET',
		path: '/api/spec/storageprofile', 
		ip: null,                                                                                                                                    
		port: null, 
		type: null,
		func: null,                                                                                
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                            
		continue: 'next',                                                                          
		name: 'get_create',
		verb: 'GET',
		path: null,                                                                                
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,   
		embedOptions: null,                                                                            
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	},
];

var funcStepsDelete = [                                                                          
	{                                                                                              
			continue: 'next',                                                                          
			name: 'del_getpolicies',
			verb: 'GET',
			path: '/api/endeavour/policy?filter=[{"property":"name","value":"qqVVqq"}]', 
			ip: null,                                                                                                                                     
			port: null, 
			type: null,
			func: null,                                                                                 
			cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          
		name: 'del_delpolicies',
		verb: 'GET',
		//path: '/api/spec/storageprofile?filter=[{"property":"name","value":"qqVVqq"}]',                  // NB: not yet supported in SPP
		path: '/api/spec/storageprofile', 
		ip: null,                                                                                                                                     
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ logMsg('info', 'Step=' + this.name + ' finished');}
},  	    
];


const getSlapStepFuncs = {
	get_issprof: function(step, funcStream, ecxSessionId, oFromNg, cb){                                  // resolving /ecxngp/slapolicy implies /api/spec/storageprofile (at least 1) must exist at backend
		ngp.isSprof(step, funcStream, ecxSessionId, oFromNg, cb);   
	},
	get_create: function(step, funcStream, ecxSessionId, oFromNg, cb){                                   // this step will process any returned sprofs
		genSlaps(step, funcStream, ecxSessionId, oFromNg, cb);    
	},	
	del_getpolicies: function(step, funcStream, ecxSessionId, oFromNg, cb){                              
		getPolicies(step, funcStream, ecxSessionId, oFromNg, cb);    
	},
	del_delpolicies: function(step, funcStream, ecxSessionId, oFromNg, cb){                              
		delPolicies(step, funcStream, ecxSessionId, oFromNg, cb);    
	},			
};

exports.init = init;
var HashMap = require('hashmap');

function init(req, res, subType, slaPolicyName, embedSLAOptions)                                                                       // presence of subType means URL contained status=true&subtype=<subtype>
{
	var ngState = ngp.getSessionState(req, res);
	
	if ( ngState.stale )                                                                                 // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                            // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var funcSteps = null;
	var genericFuncName = null;

	if ( req.method == 'GET' )
	{
			genericFuncName = 'get_slapolicy';
			funcSteps = JSON.parse(JSON.stringify(funcStepsGet));
	}   

	if ( req.method == 'DELETE' )
	{
			genericFuncName = 'delete_slapolicy';
			funcSteps = JSON.parse(JSON.stringify(funcStepsDelete));
	}   
	
	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	stepsStream.forEach(function(step){
		step.ip = ngp.ecxSidHash[ecxSid].host;
		step.port = ngp.ecxExternal;

		if ( step.name == 'get_create' )
		{
			step.embedOptions = embedSLAOptions;
		}
		
		if ( step.name == 'get_issprof' )
		{
			var appendToReqUrl = req.url.split("?");			
			step.path = step.path + "?" + appendToReqUrl[1];			//For Pagination for SLA Policies
		}
		
 		if ( step.name == 'del_getpolicies' )
		 step.path = step.path.replace('qqVVqq', '*' + require('querystring').escape(slaPolicyName));
		
		if ( step.name == 'del_delpolicies' )
			step.path = step.path + '::' + require('querystring').escape(slaPolicyName);                                                 // bit of a hack, eh?

		step.func = getSlapStepFuncs[step.name];                                                   
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};	
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, genericFuncName);                                  // add reqId to steps, create store for the req & set currentFunc to genericFuncName

	ngState.currentFunc = genericFuncName;                                                               
	//@#@ngp.setEndpointState(ecxSid, 'statusSpec', subType);                                                 // MUST do after setting currentfunc - used by GET logic only
	ngp.setRequestState(req.reqId, ecxSid, 'statusSpec', subType);
	return stepsStream;
}


/*
  Generate SLA Policies is called to return generate a so-called SLA Policy. A "slap" is effectively equivalent to the original storage profile object POSTed by the GUI. Consequently, this 
  function can be viewed as the inverse of POST /api/storageprofile.

  From another perspective, whereas the POST /api/storageprofile "splits" the POST body into multiple sub-policies, GET /api/slapolicy "merges" the multiple sub-policies back into the 
  /api/storageprofile POST body.
*/      

var genSlaps = function(step, funcStream, ecxSessionId, oFromNg, cb) {   
	//@#@var compositeRo = ngp.getEndpointState(ecxSessionId, 'sprofs');   
	var compositeRo = ngp.getRequestState(step.reqId, ecxSessionId, 'sprofs');       	
	var sprofs = compositeRo.rO;

	var slaPoliciesAr = [];
	var sprofClones = JSON.parse(JSON.stringify(sprofs.storageprofiles));                                //!! clone not needed; this is list of sprofs from backend

	ngp.ngpCoreHelper.getSites(ecxSessionId, function(getSitesResObj) {
		ngp.ngpCoreHelper.getPolicies(ecxSessionId, function(getPoliciesResObj) {
			var policiesArray= getPoliciesResObj.rO['policies'];		
	    	var arrayOfSites = getSitesResObj.rO['sites'];
		    var mapSitesNameId = new HashMap();
					
		    for (var i = 0; i < arrayOfSites.length; i++) {
				mapSitesNameId.set(arrayOfSites[i]['id'],arrayOfSites[i]['name']);
		    }

		    mergeSprofToSlap(step, sprofClones, slaPoliciesAr, ecxSessionId, mapSitesNameId, policiesArray, 'ok', function(rc) {
				if ( rc != 'ok' ) {
					var failResult = ngp.getGenericEcxErrorResponse(ngp.getRequestState(step.reqId, ecxSessionId, 'addStatusErrorResponse'), ecxSessionId, step);			
			    	ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult);                  // drive steps to fail
					return;
				}

				// FTI each storage profile has been inversed to its original POST /ecxngp/storageprofile form. 

				var rtrnSlapO = {                                                                                  // accum all the slaps we generated; this goes back to NG GUI
					links: sprofs.links,
					total: sprofs.total,
					slapolicies: slaPoliciesAr,
				};

				var successResult = {
					statusCode: 200,
					response: rtrnSlapO,
				};

				ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, successResult);	             
			});
		});
	});
};

function mergeSprofToSlap(step, sprofClones, slaPoliciesAr, ecxSid, mapSitesNameId, policiesArray, lastRc, cb) {
	if ( sprofClones.length < 1 ) {
		cb(lastRc); 
		return;
	}

	var sprof = sprofClones.shift();
	if ( ! sprof.spec.simple ) {
		sprof.spec.simple = false;
		slaPoliciesAr.push(sprof);
		ngp.logMsg('debug', 'Ignoring non-simple storage profile');                                        
		mergeSprofToSlap(step, sprofClones, slaPoliciesAr, mapSitesNameId, policiesArray, ecxSid, lastRc, cb);                            // RECURSE; we've no truck with non-simple storage profiles
		return;
	}
  
	var slapAr = [];
	var subpHash = ngp.arrayToHash(sprof.spec.subpolicy, 'type');                                        // NB: *assumes* no more than 1 of SNAPSHOT, REPLICATION, BACKUP

    if ( subpHash['SNAPSHOT'] ) {
		slapAr.push({
			type: 'SNAPSHOT',
			retention: subpHash.SNAPSHOT.spec.option.retention,                                              // tolerate .age or .numsnapshots
			trigger: subpHash.SNAPSHOT.trigger,
			label: subpHash.SNAPSHOT.label
	    });
	}

	if ( subpHash['SP'] ) {		
		var slapHappy = {
			type: 'SP',
		    trigger: subpHash.SP.trigger,
	    };

		if ( subpHash.SP.useVsnap )
		slapHappy.useVsnap = subpHash.SP.useVsnap;

		if ( subpHash.SP.destination )
		slapHappy.destination = subpHash.SP.destination;
	    slapAr.push(slapHappy);
	}

    if ( subpHash['REPLICATION'] ) {
		slapAr.push({
			type: 'REPLICATION',
			software: false,
			retention: subpHash.REPLICATION.spec.destination.option.retention,
            useEncryption: subpHash.REPLICATION.spec.destination.option.useEncryption || false,
			trigger: subpHash.REPLICATION.trigger,
			site: mapSitesNameId.get(subpHash.REPLICATION.spec.destination.target.id),
		});	
	}

	if ( subpHash['SPPOFFLOAD'] ) {
		slapAr.push({
			type: 'SPPOFFLOAD',
			retention: subpHash.SPPOFFLOAD.spec.destination.option.retention, 
			target: subpHash.SPPOFFLOAD.spec.destination.target,     
			trigger: subpHash.SPPOFFLOAD.trigger,               
			source: subpHash.SPPOFFLOAD.source,
		});
	}

	if ( subpHash['SPPARCHIVE'] ) {
		slapAr.push({
			type: 'SPPARCHIVE',
			retention: subpHash.SPPARCHIVE.spec.destination.option.retention, 
			target: subpHash.SPPARCHIVE.spec.destination.target,     
			trigger: subpHash.SPPARCHIVE.trigger,               
			source: subpHash.SPPARCHIVE.source,
		});
	}

	if ( subpHash['BACKUP'] ) {
		slapAr.push({
			type: 'REPLICATION',
			retention: subpHash.BACKUP.spec.option.retention,
            useEncryption: subpHash.BACKUP.spec.destination.option ? subpHash.BACKUP.spec.destination.option.useEncryption || false : false,
			software: true,
			trigger: subpHash.BACKUP.trigger,                                                                //@@ sprof - the source property may not exist (old stuff)!!
			site: mapSitesNameId.get(subpHash.BACKUP.spec.destination.target.id),
		});		
	}

	var modifiedLinks = modLinks(ecxSid, sprof.links, sprof.name);
		
	var slap = {
		id: sprof.id,
		links: modifiedLinks,
		description: sprof.description,
		name: sprof.name,
		version: '1.0',
		spec: {
			simple: sprof.spec.simple,                                                                   
			subpolicy: slapAr,                                                                               // point to subpolicies that we modified/created; see note 137
		},
		rbacPath: sprof.rbacPath,  
	};

	if(sprof.spec.demo)
	slap.spec.demo = sprof.spec.demo;

	slaPoliciesAr.push(slap);					
	mergeSprofToSlap(step, sprofClones, slaPoliciesAr, ecxSid, mapSitesNameId, policiesArray, lastRc, cb);                          // RECURSE

}

function modLinks(ecxSid, links, slapName) {
	if ( links.delete ) {
		var urlParts = links.delete.href.split(/\:\d{3,4}/);
			
		var modifiedURLParts =  urlParts[0].split('//');
		var finalURLParts =  modifiedURLParts[0] + '//' + ngp.ecxSidHash[ecxSid].host;
			
		var escSlapName = require('querystring').escape(slapName);
		if(ngp.productionMode == false) 
		links.delete.href = finalURLParts + ':' + ngp.ngpPort + ngp.apiFirstStep + 'slapolicy?name=' + escSlapName;  // e.g. http://<IP>:8083/ecxngp/slapolicy/Gold
		else
		links.delete.href = finalURLParts +  ngp.apiFirstStep + 'slapolicy?name=' + escSlapName;  // e.g. http://<IP>:8083/ecxngp/slapolicy/Gold
	}

	return links;                                                                                      // maybe changed
}
	
//!! find and save endpoint state all policies that use the indicated storage profile. so get policy; use its job link to get the job
// save an object like {job: <ID>, policy: <ID>} 
// IN NEXT STEP YOU HAVE to GET /api/spec/storageprofile with name filter, e.g. Gold so you can get its ID and delete it

function getPolicies(step, funcStream, ecxSessionId, oFromNg, cb)
{
	var currentStepPath = step.path;
	var currentStepPathSplitForFilter = currentStepPath.split('=');
	var currentStepPathFilter = currentStepPathSplitForFilter[1];
	currentStepPathFilter = require('querystring').unescape(currentStepPathFilter);
	
	var fixedFilterOption = ngp.escape(currentStepPathFilter);
	var fixedStepPath = currentStepPathSplitForFilter[0] + '=' + fixedFilterOption;
	
	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, fixedStepPath, ecxSessionId, null, function(polO){
				var deletionList = [];

				if ( polO.e || polO.response.statusCode != 200 )        
				{
					var failResult = ngp.getGenericEcxErrorResponse(polO, ecxSessionId, step); 
					ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false);    
				}
				else
				{
					if ( polO.rO.policies.length < 1 )                                                   // FTI no policies (and thus no jobs...) belong to the storage profile in question
					{
							ngp.logMsg('warn', 'No policies reference the indicated storage profile');
							//@#@ ngp.setEndpointState(ecxSessionId, 'deleteList', deletionList); 
							ngp.setRequestState(step.reqId, ecxSessionId, 'deleteList', deletionList); 
							step.cb(ecxSessionId, null);                                                  
							ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                     // next step will delete only the sprof
					}
					else
					{
							getJobForPolicy(polO.rO.policies, deletionList, ecxSessionId, function(rc, result){
									if ( rc == 'ng' )                                                            // FTI response.e was populated, serious error
									{
											ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, result, false); 
									}
									else
									{
											//@#@ngp.setEndpointState(ecxSessionId, 'deleteList', deletionList); 
											ngp.setRequestState(step.reqId, ecxSessionId, 'deleteList', deletionList); 											
											step.cb(ecxSessionId, null);                                                  
											ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                             
									}
							});
					}
				}			
		});  	
}

function getJobForPolicy(policies, deletionList, ecxSessionId, cb)                             // recursive funk
{
		if ( policies.length < 1 )
		{
			cb('ok', null);
		}
		else
		{
			var policy = policies.shift();

			/*
			var deleteMap = {
				job: 0,
				policy: policy.id,                                                                     // !! CONSIDER re-doing this using links.delete instead
			};	
			*/

			var deleteMap = [policy.links.delete.href];
			
			ngp.sendRequestToEcx('GET', ngp.ecxExternal, policy.links.jobs.href, ecxSessionId, null, function(jobO){
				if ( jobO.e )                                                                          // only if serious error do we give up
					{
						var failResult = ngp.getGenericEcxErrorResponse(jobO, ecxSessionId);               // NB: ggeer doesn't require a 3rd step argument
						cb('ng', failResult);    
					}
					else                                                                                 // tolerate non-200 and find all policies and jobs that we can
					{
						if ( jobO.rO.jobs.length < 1 || jobO.response.statusCode != 200 )                  // FTI quelque chose bizarre: policy IS but job AIN'T (or worse if not 404)
						{
								ngp.logMsg('warn', 'Failed to find job for policy ID=' + policy.id);           // leave deleteMap incomplete (jobId of zero); we will delete policy only
								deletionList.push(deleteMap);
								getJobForPolicy(policies, deletionList, ecxSessionId, cb);
						}
						else
						{
								var job = jobO.rO.jobs[0];
								//deleteMap.job = job.id;                                                      // !! CONSIDER re-doing this using links.delete instead
								deleteMap.unshift(job.links.delete.href);                                      // NB: job before policy
								deletionList.push(deleteMap);
								getJobForPolicy(policies, deletionList, ecxSessionId, cb)                      // RECURSE
						}
					}
			});
		}
}

function delPolicies(step, funcStream, ecxSessionId, oFromNg, cb) {
	var deletionList = ngp.getRequestState(step.reqId, ecxSessionId, 'deleteList'); 
	var sprofName = require('querystring').unescape(step.path.split('::')[1]);  
	step.path = step.path.split('::')[0];

	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, null, function(sprofO) {
		if ( sprofO.e ) {
			var failResult = ngp.getGenericEcxErrorResponse(sprofO, ecxSessionId);               // NB: ggeer doesn't require a 3rd step argument
			cb('ng', failResult);    
		}
		else {
			if ( sprofO.rO.storageprofiles.length < 1 || sprofO.response.statusCode != 200 ) {
				ngp.logMsg('warn', 'Failed to find any associated SLA Policies');                       // leave deleteMap incomplete (jobId of zero); we will delete policy only
			}
			else {
				var matchSprof = null;                                                           // NB: maybe we match the sprof. If we do, we'll delete it

				sprofO.rO.storageprofiles.forEach(function(sprof) {
					if ( matchSprof )
				    return;

			     	if ( sprof.name == sprofName )
				    matchSprof = sprof;
				});

				if ( matchSprof ) {
					deletionList.push([matchSprof.links.delete.href]); 
					var resultsList = [];
								
					deleteJobsPoliciesSprof(deletionList, resultsList, ecxSessionId, function(rc, result) {
						if(rc == 'failed') {
							ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, result, true);
						}
						else {
							var result = {
								statusCode: 204,
								response: resultsList,                                                           
							};				
							ngp.setStepSuccessResult(step, funcStream, ecxSessionId, oFromNg, cb, result);		
						}                             
					});											  
				}
				else {
					ngp.logMsg('warn', "Failed to find SLA Policy : " + sprofName);
					var failResult = {
						statusCode: 404,
                        response: {rc: 'Failed to find SLA Policy : ' + sprofName}                                                                                                                             
					};
					ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);								
				}
			}
		}
	});
}

function deleteJobsPoliciesSprof(deletions, resultsList, ecxSessionId, cb) {
	if ( deletions.length < 1 ) {
		cb('ok', null);
	}
	else {
		var deletion = deletions.shift();                                                        // and deletion is an array of one or two elements

        // NB If job is present then it'll be deleted in outer loop, delete it in outer loop and policy in inner. BUT IF NO JOB, then delete policy in outer and do nothing in inner!

		ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, deletion[0], ecxSessionId, null, function(delO) {
			if ( delO.e ) {                                                                          // only if serious error do we give up
				//	var failResult = ngp.getGenericEcxErrorResponse(delO, ecxSessionId);             // NB: ggeer doesn't require a 3rd step argument
				//	cb('ng', failResult);    
				ngp.logMsg('warn', "Continuing after outer deletion error=" + delO.e + ' against entity=' + deletion[0]);
			}
			else {
				if ( delO.response.statusCode == 204 ) {
					resultsList.push(deletion[0]);
					deletion.shift();

					if ( deletion.length < 1 )                                                         // FTI done with single element (policy | sprof) deletion array
					deleteJobsPoliciesSprof(deletions, resultsList, ecxSessionId, cb);             // RECURSE
					else {                                                                              // EHI about to finish last (policy) element in a 2 element deletion array
						ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, deletion[0], ecxSessionId, null, function(delO) {
							if ( delO.e ) {                                                                          // only if serious error do we give up
								ngp.logMsg('warn', "Continuing after inner deletion error=" + delO.e + ' against entity=' + deletion[0]);
							}
							else {
								if ( delO.response.statusCode == 204 ) {
									resultsList.push(deletion[0]);
								}
								else {
									ngp.logMsg('warn', 'Failed to delete ' + deletion[0]);           
								}
							}
							deleteJobsPoliciesSprof(deletions, resultsList, ecxSessionId, cb);         // RECURSE
						});
					}
				}
				else {
					ngp.logMsg('warn', 'Failed to delete ' + deletion[0]);    
					var failResult = {
						statusCode: delO.response.statusCode,
						response: {rc: (JSON.stringify(delO.rO.description).replace(/\"/g, ""))}                                                                                                                           
					};  
					cb('failed', failResult);     
				}			
			}
		});
	}
}