if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');                                                                           


var funcSteps = [                                                                                  
	{                                                                                             
		continue: 'next',                                                                          
		name: 'reg_usermode',
		verb: 'GET',
		path: '/api/identity/user',
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'reg_createuser',   
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: this block is changed before registration to conform to entity being registered
		continue: 'next',         
		name: 'reg_register', 
		verb: 'POST',    
		path: null,                                                                                // NB: depends on type of entity we are registering  
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set to skip if reg_register succeeds, but if driven it will delete identity only if we created one  
		name: 'reg_deleteuser',                                                                    
		verb: 'DELETE',    
		path: null,   
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 	
];


const regProviderStepFuncs = {
	reg_usermode: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		if ( ! oFromNg.username )                                                                  // smtp reg don't require user, ECC-10990
		{
			step.cb(ecxSessionId, null);                                                                 
			funcStream[0].continue = 'skip';                                                       // zap create user step
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                           
		}
		else
			ngp.userMode(step, funcStream, ecxSessionId, oFromNg, cb);
	},
	reg_createuser: function(step, funcStream, ecxSessionId, oFromNg, cb){                         // NB: no failsafe cuz nothing done prior to this step
		ngp.createUser(step, funcStream, ecxSessionId, oFromNg, cb); 
	},
	reg_register: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		register(step, funcStream, ecxSessionId, oFromNg, cb);		
	},
	reg_deleteuser: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		ngp.regDeleteUser(step, funcStream, ecxSessionId, oFromNg, cb);
	},			
};


exports.init = init;


function init(req, res, providerType)                                    
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                                    // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                          

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var typology = ngp.getRegistrationTypology(req, providerType);                                              

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	var regStep = null;

	stepsStream.forEach(function(step){                                                                     // find the registration step
		if ( regStep )
			return;

		if ( step.name == 'reg_register' )
		{
			regStep = step;
		}
	});

	regStep.path = typology.path;                                                                           // NB: proper URL path against which to POST

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.resourceType = typology.resourceType;
		step.type = typology.type;                                                                          // NB: we need to be told the API path for type we are registering

		step.func = regProviderStepFuncs[step.name];                                                        // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'register_' + typology.resourceType + '_' + typology.type);  

	ngState.currentFunc = 'register_' + typology.resourceType + '_' + typology.type;                        // e.g., register_storage_netapp

	return stepsStream;
}


var register = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
	var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);
	ngp.logMsg('info', "Input from client for SMTP or LDAP registration in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));

	var ngState = ngp.ecxSidHash[ecxSessionId];

	//var postBody = getProviderRegistrationPostBody(ecxSessionId, step, ngState, oFromNg);
	var postBody = getProviderRegistrationPostBody(step, ngState, oFromNg);
	//!! UNCOMMENT to force registration failure:  postBody.hostAddress = 'godFatherOfSoul';

	ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(regO){ 
		if ( regO.e || regO.response.statusCode != 201 )   
		{
			var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 

			//@#@var createdIdentities = ngp.getEndpointState(ecxSessionId, 'createdIdentities'); 
			var createdIdentities = ngp.getRequestState(step.reqId, ecxSessionId, 'createdIdentities'); 

			if ( ( ! createdIdentities || createdIdentities.length < 1 ) && funcStream.length > 0 )         // FTI we never created an identity AND subsequent step is present                                  
				funcStream[0].continue = 'skip'; 

			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);           // boolean arg causes ssfr not to set 'fail' (we've already set 'skip')
		}
		else
		{
			ngp.logMsg('debug', "Registration OK response=" + JSON.stringify(regO.rO));  
	
			var successResult = {
				statusCode: regO.response.statusCode,
				response: regO.rO,
			};

			ngp.setSerialStreamResult(ecxSessionId, successResult);                                         // NB: let caller see result of successful update

			if ( funcStream.length > 0 )                                                                    // FTI there is a next step (always first element in what remains of array)
			{
				funcStream[0].continue = 'skip';                                                            // cause skip of next serial step (deletion of identity)                        
			}

			step.cb(ecxSessionId, regO.rO);                                                  
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 					
		
		}
	});
};


//function getProviderRegistrationPostBody(ecxSessionId, step, ngState, oFromNg)                 // called by regRegister 
function getProviderRegistrationPostBody(step, ngState, oFromNg)                                            // called by regRegister 
{
	var postBody = null;

	postBody = {  
		name: oFromNg.name,
		hostAddress: oFromNg.hostAddress,                                                             
		portNumber: oFromNg.portNumber,                                                                     // use whatever GUI gives us, if anything; else backend decides
		sslConnection: oFromNg.sslConnection,                                                               // ditto
		
		user: {
			//@#@href: ngp.getEndpointState(ngState.ecxSid, 'userHref'),                                         // may be href: null if no username given by NG
			href: ngp.getRequestState(step.reqId, ngState.ecxSid, 'userHref'),                                         // may be href: null if no username given by NG
		},

		name: oFromNg.hostAddress,
	};


// Handle DIFFERENTIAE here. 

	if ( step.resourceType == 'smtp' )
	{
		delete postBody.type;
		
		postBody.timeout = oFromNg.timeout;
		postBody.fromAddress = oFromNg.fromAddress;
		postBody.subjectPrefix = oFromNg.subjectPrefix;

		if ( ! oFromNg.username )                                                                           // ECC-10990
		{
			delete postBody.user;
		}
	}

	if ( step.resourceType == 'ldap' )
	{
		delete postBody.type;
		
		postBody.baseDN = oFromNg.baseDN;
		postBody.userFilter = oFromNg.userFilter;
		postBody.groupRDN = oFromNg.groupRDN;
		postBody.userRDN = oFromNg.userRDN;
	}

	return postBody;
}
